package jp.sagalab.skit;

import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.Vector;

/**
 * 絶対図形空間操作のコントローラ。
 * @author nagase
 */
public class AbsoluteModelSpaceController {
	public static AbsoluteModelSpaceController create( ReadWriteSKITModel _model ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		return new AbsoluteModelSpaceController(_model );
	}

	/**
	 * 指定したそれぞれ点を用いて、絶対的図形空間操作を行う。
	 * 各点の座標はモデル空間上の座標値である。
	 * @param _pibot 固定点。
	 * @param _before 移動前の点。
	 * @param _after 移動後の点。
	 */
	public void move(Point _pibot,Point _before,Point _after){
		// モデル空間への変換行列に掛ける行列を算出
		TransformMatrix transformMatrix = calcTransformMatrix( _pibot, _before,_after );
		if ( transformMatrix == null ) {
			// 拡大縮小率が不定の場合、何もしない
			// XXX 操作不能になるため、救済措置が必要か？
			return;
		}
		TransformMatrix tmp = m_model.getViewToModel();
		if ( updateTransform( transformMatrix ) ) {
			if ( !updateGridSpace( transformMatrix ) ) {
				//グリッド空間の更新に失敗した場合モデル空間を元の状態に戻す
				m_model.setViewToModel( tmp );
			}
		}
	}

	/**
	 * 指絶対的図形空間操作を行うための変換行列を求める。
	 * _pibotを中心として_beforeを_afterに移動させる相似変換を行う変換行列を返す。
	 * @param _pibot 固定点。
	 * @param _before 移動前の点。
	 * @param _after 移動後の点。
	 * @return 変換行列。
	 */
	private static TransformMatrix calcTransformMatrix( Point _pivot, Point _before, Point _after ) {
		Vector beforeVector = Vector.createSE( _pivot, _before );
		Vector afterVector = Vector.createSE( _pivot, _after );

		// 拡大縮小率を算出
		double scalingRatio = beforeVector.length() / afterVector.length();
		if ( Double.isNaN( scalingRatio ) || Double.isInfinite( scalingRatio ) ) {
			// 拡大縮小率が不定の場合
			return null;
		}

		// 拡大縮小の変換行列を生成
		TransformMatrix scaling = TransformMatrix.scaling( scalingRatio );

		// 回転角
		double angle = afterVector.angle( beforeVector );
		// 回転軸
		Vector axis = afterVector.cross( beforeVector );
		// 回転行列を生成
		TransformMatrix rotation = TransformMatrix.rotation( angle, axis );

		// 原点を固定点の位置へ平行移動させる変換行列を生成
		TransformMatrix translation = TransformMatrix.translation( _pivot.x(), _pivot.y(), _pivot.z() );
		// 固定点を原点へ平行移動させる変換行列を生成
		TransformMatrix reversedTranslation = TransformMatrix.translation( -_pivot.x(), -_pivot.y(), -_pivot.z() );

		// 変換行列
		TransformMatrix transformMatrix = reversedTranslation.product( rotation ).product( scaling ).product( translation );

		return transformMatrix;
	}
	/**
	 * モデル空間への変換行列を更新する。
	 *
	 * @param _mat モデル空間上での平行移動・スケーリング・回転を表す変換行列。
	 */
	private boolean updateTransform( TransformMatrix _mat ) {
		TransformMatrix transform = m_model.getViewToModel().product( _mat );
		double min = 1 / m_scale;
		double max = m_scale;
		if ( transform.scalalize() >= min && transform.scalalize() <= max ) {
			m_model.setViewToModel( transform );
			return true;
		}
		return false;
	}

	/**
	 * グリッド空間を更新する。
	 *
	 * @param _mat モデル空間での平行移動・スケーリング・回転を表す変換行列。
	 */
	private boolean updateGridSpace( TransformMatrix _mat ) {
		double scalalize = _mat.scalalize();
		GridSpace gridSpace = m_model.getGridSpace();
		TransformMatrix mat = m_model.getGridSpace().grid().product( _mat );
		gridSpace = GridSpace.create(mat,
			scalalize * gridSpace.basedFuzziness(), gridSpace.factor() );
		double min = 1 / m_scale * 1e-3;
		double max = m_scale * 1e3;
		if ( mat.scalalize() >= min && mat.scalalize() <= max ) {
			m_model.setGridSpace( gridSpace );
			return true;
		}
		return false;
	}

	/** モデル */
	private final ReadWriteSKITModel m_model;
	/** 拡縮倍率 XXX 2の10乗に決め打ち */
	private static final double m_scale=1024;
	/**
	 * コンストラクタ.
	 *
	 * @param _model 読み書き可能なSKITモデル
	 */
	private AbsoluteModelSpaceController( ReadWriteSKITModel _model ) {
		m_model=_model;
	}
}
