package jp.sagalab.skit;


import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;

/**
 * 入力デバイスを扱うための抽象骨格実装．
 *
 * @author fujinaga
 * @since 2008/10/14
 */
public abstract class AbstractSerialInputDevice {
	/** シリアルポート */
	protected SerialPort m_serialPort;

	/** デバイスから情報を入力するためのストリーム */
	protected InputStream m_inputStream;

	/** デバイスに情報を出力するためのストリーム */
	protected OutputStream m_outputStream;

	/**
	 * 通信する際のボーレート．
	 */
	public static enum BAUD_RATE {
		VALUE_1200, VALUE_2400, VALUE_4800, VALUE_9600, VALUE_19200, VALUE_38400, VALUE_57600, VALUE_115200, VALUE_230400, VALUE_460800;
		/**
		 * BAUD_RATE型の要素を，int型の対応した値に変換する．
		 *
		 * @return int型の対応したボーレート値．
		 */
		public int toInt() {
			switch ( this ) {
			case VALUE_1200:
				return 1200;
			case VALUE_2400:
				return 2400;
			case VALUE_4800:
				return 4800;
			case VALUE_9600:
				return 9600;
			case VALUE_19200:
				return 19200;
			case VALUE_38400:
				return 38400;
			case VALUE_57600:
				return 57600;
			case VALUE_115200:
				return 115200;
			case VALUE_230400:
				return 230400;
			case VALUE_460800:
				return 460800;
			default:
				return 1200;
			}
		}
	}

	/**
	 * シリアルポートを開く．
	 *
	 * @param _portName
	 *          シリアルポートの名前．
	 * @param _baudrate
	 *          ボーレート．
	 * @throws NoSuchPortException
	 *           指定した名前のポートが見つからないとき．
	 * @throws PortInUseException
	 *           指定したポートが他のアプリケーションで使用中のとき．
	 * @throws UnsupportedCommOperationException
	 *           指定したパラメータが不正な値のとき．
	 * @throws IOException
	 *           入出力ストリームが開けないとき．
	 */
	protected void open( String _portName, BAUD_RATE _baudrate ) throws NoSuchPortException, PortInUseException,
			UnsupportedCommOperationException, IOException {

		try {
			// 指定した名前のポート識別子を取得して開く．3000はタイムアウトまでの時間[ミリ秒]．
			CommPortIdentifier portID = CommPortIdentifier.getPortIdentifier ( _portName );
			m_serialPort = ( SerialPort ) portID.open ( _portName, 3000 );
			m_serialPort.setSerialPortParams ( _baudrate.toInt (), SerialPort.DATABITS_8, SerialPort.STOPBITS_2,
					SerialPort.PARITY_NONE );
			m_inputStream = m_serialPort.getInputStream ();
			m_outputStream = m_serialPort.getOutputStream ();

		} catch ( NoSuchPortException e ) { // 指定した名前のポートが見つからないとき
			throw e;
		} catch ( PortInUseException e ) { // 指定したポートが使用中のとき
			throw e;
		} catch ( UnsupportedCommOperationException e ) { // 指定したパラメータが不正な値のとき
			throw e;
		} catch ( IOException e ) { // 入出力ストリームが開けないとき
			throw e;
		}
	}

	/**
	 * デバイスに文字列を送信する．
	 *
	 * @param _message
	 *          Fastrakに送信する文字列．
	 * @throws IOException
	 *           書き込みが失敗したとき．
	 * @return 書き込むことに成功した文字数 [Byte]．
	 */
	protected synchronized int write( String _message ) throws IOException {
		try {
			m_outputStream.write ( _message.getBytes () );
		} catch ( IOException e ) {
			throw e;
		}
		return _message.length ();
	}

	/**
	 * デバイスからデータを読み込む．
	 *
	 * @param _dataByte
	 *          読み込むデータのサイズ [Byte]．
	 * @throws IOException
	 *           読み込みが失敗したとき ．
	 * @return Fastrakから得られたデータ．
	 */
	protected synchronized String read( int _dataByte ) throws IOException {
		long startTime = System.currentTimeMillis (); // 読み込み開始時刻
		long spendTime = 0; // 読み込み開始からの経過時間
		final long TIMEOUT = 3000; // タイムアウトまでの時間[ミリ秒]
		int readByte = 0; // 読み込んだバイト数
		byte[] buffer = new byte[ _dataByte ]; // 読み込みバッファ

		try {
			while ( readByte < _dataByte && spendTime < TIMEOUT ) {
				readByte = m_inputStream.read ( buffer );
				spendTime = System.currentTimeMillis () - startTime;
			}
		} catch ( IOException e ) {
			throw e;
		}
		return new String ( buffer, 0, readByte );
	}


	protected synchronized String readLine() throws IOException{
		BufferedReader bufferReader = new BufferedReader(new InputStreamReader(m_inputStream));
		return bufferReader.readLine();
	}

}
