package jp.sagalab.skit;

import java.awt.AWTException;
import java.awt.GraphicsEnvironment;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.event.InputEvent;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import jp.sagalab.jflib.ValidateUtils;
import jp.sagalab.jftk.curve.BezierCurve;
import jp.sagalab.jftk.curve.primitive.Line;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.curve.Range;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.curve.ParametricEvaluable.EvaluationType;
import jp.sagalab.jftk.curve.primitive.CircularArc;
import jp.sagalab.jftk.curve.primitive.EllipticArc;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * ダミークライアント。
 */
public class DummyClient implements Runnable {
	/** サーバに接続しているソケット */
	private final Socket m_socket;

	/**
	 * 生成。
	 * @param _socket サーバに接続済みのソケット。
	 */
	public DummyClient( Socket _socket ) {
		// ダミーなのでエラーチェックは省略
		m_socket = _socket;
	}

	/**
	 * ディスプレイ全体をビューポートとするダミーXML文書をサーバへ送信する。
	 * @param _grid グリッド幅。
	 * @throws IOException 入出力エラーが発生したとき。
	 * @throws IllegalArgumentException _gridがnullであるか、
	 * _gridの要素に0や負数が含まれるとき。
	 */
	public void sendDummyInfo( double _grid ) throws IOException {
		// エラーチェック
		ValidateUtils.throwExceptionIfNull( _grid, "_grid" );
		double grid = _grid;
		if ( grid <= 0 ) {
			throw new IllegalArgumentException( "_grid must not contain 0 or negative value" );
		}

		// ダミーXML文書を生成
		String dummyString = createDummyString( grid );

		// XML文書をソケットに送信
		BufferedOutputStream out = new BufferedOutputStream( m_socket.getOutputStream(), m_socket.getSendBufferSize() );
		out.write( dummyString.getBytes() );
		out.flush();
	}

	/**
	 * ダミーのXML文書を生成して返す。
	 * ビューポートはディスプレイ解像度全体、モデル空間への変換行列は単位行列が用いられる。
	 * @param _grid グリッド幅。
	 * @return 指定したグリッドサイズ列をもつダミーXML文書。
	 */
	private static String createDummyString( double _grid ) {
		// ビューポート情報の構築
		GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
		Rectangle bounds = env.getMaximumWindowBounds();
		String viewportInfo = createViewportInfo( bounds.x, bounds.y, bounds.width, bounds.height );

		// モデル空間情報の構築
		String modelSpaceInfo = createModelSpaceInfo( _grid );

		return "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>"
			+ "<skit>" + viewportInfo + modelSpaceInfo + "</skit>\n";
	}

	/**
	 * ビューポート情報を表現するXML文書を生成して返す。
	 * @param _x ビューポート情報のx属性値。
	 * @param _y ビューポート情報のy属性値。
	 * @param _width ビューポート情報のwidth属性値。
	 * @param _height ビューポート情報のheight属性値。
	 * @return
	 */
	private static String createViewportInfo( int _x, int _y, int _width, int _height ) {
		StringBuilder builder = new StringBuilder( 128 );
		builder.append( "<viewport_information" );
		builder.append( " height=" );
		builder.append( "\"" ).append( _height ).append( "\"" );
		builder.append( " width=" );
		builder.append( "\"" ).append( _width ).append( "\"" );
		builder.append( " x=" );
		builder.append( "\"" ).append( _x ).append( "\"" );
		builder.append( " y=" );
		builder.append( "\"" ).append( _y ).append( "\"" );
		builder.append( " />" );
		return builder.toString();
	}

	/**
	 * モデル空間情報を表現するXML文を生成して返す。
	 * @param _grid グリッド幅。
	 * @return モデル空間情報を表現するXML文。
	 */
	private static String createModelSpaceInfo( double _grid ) {
		StringBuilder builder = new StringBuilder( 256 );
		builder.append( "<model_space_information>" );
		builder.append( createMappingMatrix() );
		builder.append( createGridSpaceMatrix( _grid ) );
		builder.append( "</model_space_information>" );
		return builder.toString();
	}

	/**
	 * マッピング情報を表現するXML文を生成して返す。
	 * @return マッピング情報を表現するXML文。
	 */
	private static String createMappingMatrix() {
		StringBuilder builder = new StringBuilder( 128 );
		builder.append( "<viewport_space_to_model_space_mapping_matrix" );
		for ( int i = 0; i < 4; ++i ) {
			for ( int j = 0; j < 4; ++j ) {
				double value = ( i == j ? 1 : 0 );
				builder.append( " m" ).append( i ).append( j );
				builder.append( "=\"" ).append( value );
				builder.append( "\" " );
			}
		}
		builder.append( " />" );
		return builder.toString();
	}

	/**
	 * グリッド空間情報を表現するXML文を生成して返す。
	 * @param _gridSize グリッド幅。
	 * @return グリッド空間情報を表現するXML文。
	 */
	private static String createGridSpaceMatrix( double _gridSize ) {
		StringBuilder builder = new StringBuilder( 128 );
		builder.append( "<snapping_information>" );
		builder.append( "<unit_grid_space_to_model_space_mapping_matrix" );
		for ( int i = 0; i < 4; ++i ) {
			for ( int j = 0; j < 4; ++j ) {
				double value = ( i == j ? _gridSize : 0 );
				if ( i == 3 && j == 3 ) {
					builder.append( " m" ).append( i ).append( j );
					builder.append( "=\"" ).append( 1.0 );
					builder.append( "\" " );
				} else {
					builder.append( " m" ).append( i ).append( j );
					builder.append( "=\"" ).append( value );
					builder.append( "\" " );
				}
			}
		}
		builder.append( " />" );
		// 派生倍率
		builder.append( "<derivational_scale_factor>" ).append( 2 ).append( "</derivational_scale_factor>" );
		builder.append( "</snapping_information>" );
		return builder.toString();
	}

	/**
	 * SKITからXML文書を受信し、標準出力へ出力する。
	 */
	@Override
	public void run() {
		try {
			// クライアントからのデータを受信
			String str = new BufferedReader( new InputStreamReader( m_socket.getInputStream() ) ).readLine();
			System.out.println( "output!" );
			// 受信したXML文書を解析
			Document doc;
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			doc = builder.parse( new InputSource( new StringReader( str ) ) );
			readDocument( doc );

//			BufferedInputStream in = new BufferedInputStream( m_socket.getInputStream(), m_socket.getReceiveBufferSize() );
//			byte[] buf = new byte[ 256 ];
//			while ( true ) {
//				int count = in.read( buf );
//				if ( count == -1 ) {
//					// 切断されたので接続を閉じる
//					m_socket.close();
//					return;
//				} else {
//					// 受信したXML文書を、適当な場所で改行しながら標準出力へ出力
//					for ( int i = 0; i < count; ++i ) {
//						char c = (char) buf[i];
//						System.out.print( c );
//						if ( c == '>' ) {
//							System.out.println( "" );
//						}
//					}
//				}
//			}
		} catch ( SAXException ex ) {
			Logger.getLogger( DummyClient.class.getName() ).log( Level.SEVERE, null, ex );
		} catch ( ParserConfigurationException ex ) {
			Logger.getLogger( DummyClient.class.getName() ).log( Level.SEVERE, null, ex );
		} catch ( IOException ex ) {
			Logger.getLogger( DummyClient.class.getName() ).log( Level.SEVERE, null, ex );
			try {
				m_socket.close();
			} catch ( IOException ignored ) {
			}
		}
	}

	/**
	 * SKITのドキュメントを解析し，要素に応じてマウスモジュールを起動する
	 * @param _doc 受信したSKITのXMLデータ
	 */
	private static void readDocument( Document _doc ) {
		// SKITノードの取得
		NodeList skitNode = _doc.getElementsByTagName( "skit" );
		// shapesノード群の取得
		NodeList shapesNode = skitNode.item( 0 ).getChildNodes();
		if ( shapesNode.getLength() <= 0 ) {
			return;
		}
		// shapeノード群の取得
		NodeList shapeNode = shapesNode.item( 0 ).getChildNodes();
		// shapeノード群の数
		int shapeLength = shapeNode.getLength();

		//ビューポートの作成
		GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
		Rectangle bounds = env.getMaximumWindowBounds();
		TransformMatrix viewport = TransformMatrix.translation( bounds.x, bounds.y, 0);

		for ( int i = 0; i < shapeLength; ++i ) {
			// 曲線ノード群の取得
			NodeList primitiveNode = shapeNode.item( i ).getChildNodes();
			// 曲線ノード群の数
			int primitiveNum = primitiveNode.getLength();
			// 曲線ノードの種類に応じてマウスモジュールを起動
			for ( int j = 0; j < primitiveNum; ++j ) {
				Element primitive = (Element) primitiveNode.item( j );
				System.out.println( primitive.getTagName() );

				if ( primitive.getTagName().matches( "line" ) ) {
					moveLine( primitive.getChildNodes(), viewport );
				} else if ( primitive.getTagName().matches( "circle" ) ) {
					moveCircle( primitive.getChildNodes(), true, viewport );
				} else if ( primitive.getTagName().matches( "circular_arc" ) ) {
					moveCircle( primitive.getChildNodes(), false, viewport );
				} else if ( primitive.getTagName().matches( "ellipse" ) ) {
					moveEllipse( primitive.getChildNodes(), true, viewport );
				} else if ( primitive.getTagName().matches( "elliptic_arc" ) ) {
					moveEllipse( primitive.getChildNodes(), false, viewport );
				} else if ( primitive.getTagName().matches( "open_free_curve" ) ) {
					moveFreeCurve( primitive.getChildNodes(), false, viewport );
				} else if ( primitive.getTagName().matches( "closed_free_curve" ) ) {
					moveFreeCurve( primitive.getChildNodes(), true, viewport );
				}
			}
		}
	}

	/**
	 * line要素に記述された値を用いてマウスを動作させる
	 * @param _nodes line要素のノード
	 * @param _viewport 画面のビューポート
	 */
	private static void moveLine( NodeList _nodes, TransformMatrix _viewport ) {
		// 始点をつくる
		Element el = (Element) _nodes.item( 0 );
		Point start = Point.createXYZ( Double.valueOf( el.getAttribute( "x" ) ),
			Double.valueOf( el.getAttribute( "y" ) ),
			Double.valueOf( el.getAttribute( "z" ) ) );
		// 終点をつくる
		el = (Element) _nodes.item( 1 );
		Point end = Point.createXYZ( Double.valueOf( el.getAttribute( "x" ) ),
			Double.valueOf( el.getAttribute( "y" ) ),
			Double.valueOf( el.getAttribute( "z" ) ) );

		// プリミティブをつくる
		Line line = Line.create(start, end ).transform( _viewport );
		// Bezier曲線化
		BezierCurve[] beziers = line.convert();
		int num = 4;
		// 評価点列を生成する
		List<Point> evals = new ArrayList<Point>( num );
		for(BezierCurve b : beziers){
			evals.addAll( Arrays.asList( b.evaluateAll( num, EvaluationType.DISTANCE )) );
		}

		// 動かす
		runMouseModule( evals.toArray(new Point[evals.size()]) );

	}

	/**
	 * circule要素またはcircular_arc要素に記述された値を用いてマウスを動作させる
	 * @param _nodes curcleまたはcircular_arc要素のノード
	 * @param _closed 閉じているか
	 * @param _viewport 画面のビューポート
	 */
	private static void moveCircle( NodeList _nodes, boolean _closed, TransformMatrix _viewport ) {
		// 中心点をつくる
		Element el = (Element) _nodes.item( 0 );
		Point center = Point.createXYZ( Double.valueOf( el.getAttribute( "x" ) ),
			Double.valueOf( el.getAttribute( "y" ) ),
			Double.valueOf( el.getAttribute( "z" ) ) );

		// 半径を取得する
		el = (Element) _nodes.item( 1 );
		Double radius = Double.valueOf( el.getFirstChild().getNodeValue() );

		// 終了角を取得する
		el = (Element) _nodes.item( 2 );
		Double endAngle = Double.valueOf( el.getAttribute( "end" ) );

		// 姿勢行列を取得する
		el = (Element) _nodes.item( 3 );
		double[][] element = new double[ 4 ][ 4 ];
		for ( int i = 0; i < element.length; ++i ) {
			for ( int j = 0; j < element[0].length; ++j ) {
				element[i][j] = Double.valueOf( el.getAttribute( "m" + i + j ) );
			}
		}
		TransformMatrix posture = TransformMatrix.create(element );

		// プリミティブをつくる
		CircularArc circularArc = CircularArc.create(center, radius, endAngle, posture ).transform( _viewport );
		// Bezier曲線化
		BezierCurve[] beziers = circularArc.convert();
		int num = (int)Math.round( circularArc.length()/beziers.length );
		// 評価点列を生成する
		List<Point> evals = new ArrayList<Point>( num );
		for(BezierCurve b : beziers){
			evals.addAll( Arrays.asList( b.evaluateAll( num, EvaluationType.DISTANCE )) );
		}
//		Point[] eval = circle.evaluateAll( 200, EvaluationType.DISTANCE );

		// 動かす
		runMouseModule( evals.toArray(new Point[evals.size()]) );
	}

	/**
	 * ellipse要素またはelliptic_arc要素に記述された値を用いてマウスを動作させる
	 * @param _nodes ellipseまたはelliptic_arc要素のノード
	 * @param _closed 閉じているか
	 * @param _viewport 画面のビューポート
	 */
	private static void moveEllipse( NodeList _nodes, boolean _closed, TransformMatrix _viewport ) {
		// 中心点をつくる
		Element el = (Element) _nodes.item( 0 );
		Point center = Point.createXYZ( Double.valueOf( el.getAttribute( "x" ) ),
			Double.valueOf( el.getAttribute( "y" ) ),
			Double.valueOf( el.getAttribute( "z" ) ) );

		// 長径を取得する
		el = (Element) _nodes.item( 1 );
		Double majorRadius = Double.valueOf( el.getFirstChild().getNodeValue() );

		// 短径を取得する
		el = (Element) _nodes.item( 2 );
		Double minorRadius = Double.valueOf( el.getFirstChild().getNodeValue() );

		// 開始角・終了角を取得する
		el = (Element) _nodes.item( 3 );
		Double endAngle = Double.valueOf( el.getAttribute( "end" ) );
		Double startAngle = Double.valueOf( el.getAttribute( "start" ) );

		Range range = Range.create( startAngle, endAngle );

		// 姿勢行列を取得する
		el = (Element) _nodes.item( 4 );
		double[][] element = new double[ 4 ][ 4 ];
		for ( int i = 0; i < element.length; ++i ) {
			for ( int j = 0; j < element[0].length; ++j ) {
				element[i][j] = Double.valueOf( el.getAttribute( "m" + i + j ) );
			}
		}
		TransformMatrix posture = TransformMatrix.create(element );

		// プリミティブをつくる
		EllipticArc ellipticArc = EllipticArc.create(center, majorRadius, minorRadius, range, posture ).transform( _viewport );
		// Bezier曲線化
		BezierCurve[] beziers = ellipticArc.convert();
		int num = (int)Math.round( ellipticArc.length()/beziers.length );
		// 評価点列を生成する
		List<Point> evals = new ArrayList<Point>( num );
		for(BezierCurve b : beziers){
			evals.addAll( Arrays.asList( b.evaluateAll( num, EvaluationType.DISTANCE )) );
		}
//		Point[] eval = ellipse.evaluateAll( 320, EvaluationType.DISTANCE );
		// 動かす
		runMouseModule( evals.toArray(new Point[evals.size()]) );
	}

	/**
	 * open_free_curve要素またはclosed_free_curve要素に記述された値を用いてマウスを動作させる
	 * @param _nodes open_free_curveまたはclosed_free_curve要素のノード
	 * @param _closed 閉じているか
	 * @param _viewport 画面のビューポート
	 */
	private static void moveFreeCurve( NodeList _nodes, boolean _closed, TransformMatrix _viewport ) {
		// ベジェ曲線数
		int bezierLength = _nodes.getLength();
		// 評価点列を生成する
		List<Point> evaluatedPoints = new ArrayList<Point>( bezierLength * 20 );
		for ( int i = 0; i < bezierLength; ++i ) {
			// 制御点数
			NodeList bezierNodes = _nodes.item( i ).getChildNodes();
			int controlPointLength = bezierNodes.getLength();
			Point[] controlPoints = new Point[ controlPointLength ];
			for ( int j = 0; j < controlPointLength; ++j ) {
				//要素の取得
				Element el = (Element) bezierNodes.item( j );
				controlPoints[j] = Point.createXYZ( Double.valueOf( el.getAttribute( "x" ) ),
					Double.valueOf( el.getAttribute( "y" ) ),
					Double.valueOf( el.getAttribute( "z" ) ) );
			}

			// Bezier曲線生成
			BezierCurve bezier = BezierCurve.create( controlPoints, Range.create( 0.0, 1.0 ) ).transform( _viewport );
			Point[] evaluated = bezier.evaluateAll( controlPointLength * 20, EvaluationType.DISTANCE );
			evaluatedPoints.addAll( Arrays.asList( evaluated ) );
		}
		// 動かす
		runMouseModule( evaluatedPoints.toArray( new Point[ evaluatedPoints.size() ] ) );
	}

	/**
	 * 与えられた点列に沿ってマウスをドラッグする
	 * @param _points　与えられた点列
	 */
	private static void runMouseModule( Point[] _points ) {
		try {
			Robot robot = new Robot();
			// 前のドラッグ操作終了まで一定時間空ける
			Thread.sleep( 250 );
			// マウスをドラッグ開始位置へ移動する
			robot.mouseMove( (int) _points[0].x(), (int) _points[0].y() );
			Thread.sleep( 10 );
			// マウスボタンを押す
			robot.mousePress( InputEvent.BUTTON1_MASK );

			// ボタンを押した状態で与えられた点列の位置へ移動する
			for ( int i = 0; i < _points.length; ++i ) {
				robot.mouseMove( (int) Math.round( _points[i].x() ), (int) Math.round( _points[i].y() ) );
				// 点列の個数に応じて一定時間空ける(点列が少ない状態で間隔が短いとうまく入力されない)
				// XXX C++コード見てどのくらいが適切か確認。あと割と適当なので要改善
				// Thread.sleep( 5);
				if ( _points.length < 10 ) {
					Thread.sleep( 80 );
				} else if ( _points.length > 200 ) {
					Thread.sleep( 1 );
				} else {
					Thread.sleep( 5 );
				}
			}

			// 移動が終了したらボタンを離す
			robot.mouseRelease( InputEvent.BUTTON1_MASK );
		} catch ( AWTException ex ) {
			Logger.getLogger( DummyClient.class.getName() ).log( Level.SEVERE, null, ex );
		} catch ( InterruptedException ex ) {
			Logger.getLogger( DummyClient.class.getName() ).log( Level.SEVERE, null, ex );
		}
	}
}
