package jp.sagalab.skit;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;

/**
 * 特徴点レイヤー。
 * @author oyoshi
 */
public class FeaturePointsLayer extends Component {
	private static final Color FEATURE_POINT_COLOR = new Color( 221, 155, 57, 255 );
	private static final Color PIVOT_POINT_COLOR = new Color( 215, 83, 32, 255 );
	private static final Color SELECTED_POINT_COLOR = new Color( 214, 247, 59, 255 );
	private static final Color FEATURE_POINT_SHADOW_COLOR = new Color( 63, 63, 63, 255 );
	private static final Stroke FEATURE_POINT_STROKE = new BasicStroke( 3f );
	private static final double FEATURE_POINT_RADIUS = 15.0;
	private static final double SHADOW_DISTANCE = FEATURE_POINT_RADIUS*0.1;
	private final ReadOnlySKITModel m_model;

	public static FeaturePointsLayer create( ReadOnlySKITModel _model, int _width, int _height ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		if ( _width <= 0 ) {
			throw new IllegalArgumentException( "_width <= 0" );
		}
		if ( _height <= 0 ) {
			throw new IllegalArgumentException( "_height <= 0" );
		}
		return new FeaturePointsLayer( _model, _width, _height );
	}

	private FeaturePointsLayer( ReadOnlySKITModel _model, int _width, int _height ) {
		m_model = _model;
		setBackground( new Color( 0, 0, 0, 0 ) );
		setSize( _width, _height );
	}

	public void updateScene() {
	}

	@Override
	public void paint( Graphics _g ) {
		Graphics2D g = (Graphics2D) _g;

		TransformMatrix modelToView = m_model.getViewToModel().inverse();

		Point[] featurePoints = m_model.getFeaturePoints();
		int pivotIndex = m_model.getPivotPointIndex();
		int selectedIndex = m_model.getSelectedPointIndex();
		drawFeaturePoints( g, featurePoints, pivotIndex, selectedIndex, modelToView );
	}

	/**
	 * 特徴点列を描画する。
	 * @param _g
	 * @param _featurePoints 特徴点列。
	 * @param _pivotIndex 固定点のインデックス。
	 * @param _selectedIndex 選択点のインデックス。
	 * @param _modelToView モデル空間からビューポート空間への変換行列。
	 */
	private static void drawFeaturePoints( Graphics2D _g, Point[] _featurePoints,
		int _pivotIndex, int _selectedIndex, TransformMatrix _modelToView ) {
		_g.setStroke( FEATURE_POINT_STROKE );
		_g.setColor( FEATURE_POINT_COLOR );
		for ( int i = 0; i < _featurePoints.length; ++i ) {
			if ( i != _pivotIndex && i != _selectedIndex ) {
				Point featurePoint = _featurePoints[i].transform( _modelToView );
				drawFeaturePoint( _g, featurePoint, FEATURE_POINT_COLOR );
			}
		}
		if ( 0 <= _pivotIndex && _pivotIndex < _featurePoints.length ) {
			drawPivotPoint( _g, _featurePoints[_pivotIndex].transform( _modelToView ),
				FEATURE_POINT_SHADOW_COLOR, PIVOT_POINT_COLOR );
		}
		if ( 0 <= _selectedIndex && _selectedIndex < _featurePoints.length ) {
			drawSelectedPoint( _g, _featurePoints[_selectedIndex].transform( _modelToView ),
				FEATURE_POINT_SHADOW_COLOR, SELECTED_POINT_COLOR );
		}
	}

	/**
	 * 特徴点列を描画する。
	 * @param _g
	 * @param _featurePoint
	 * @param _strokeColor
	 * @param _fillColor 
	 */
	private static void drawFeaturePoint( Graphics2D _g, Point _featurePoint, Color _fillColor ) {
		_g.setColor( _fillColor );
		SKITPanelHelper.fillCircle( _g, _featurePoint, FEATURE_POINT_RADIUS );
	}
	
	/**
	 * 固定点を描画する。
	 * @param _g
	 * @param _featurePoint
	 * @param _shadowColor
	 * @param _fillColor 
	 */
	private static void drawPivotPoint( Graphics2D _g, Point _featurePoint,
		Color _shadowColor, Color _fillColor ) {
		// 固定点の三角形計算
		double length = 50.0;
		double halfLength = length/2.0;
		Point[] points = new Point[]{
			Point.createXYZ( _featurePoint.x(), _featurePoint.y(), _featurePoint.z() ),
			Point.createXYZ( _featurePoint.x(), _featurePoint.y() - length, _featurePoint.z() ),
			Point.createXYZ( _featurePoint.x() - halfLength, _featurePoint.y() - halfLength * Math.sqrt( 3 ) , _featurePoint.z() )
		};
		Point[] shadowPoint = new Point[3];
		for( int i = 0; i < 3; ++i ){
			shadowPoint[i] = points[i].move( SHADOW_DISTANCE, SHADOW_DISTANCE, 0.0 );
		}
		// 影の描画
		_g.setColor( _shadowColor );
		SKITPanelHelper.fillPolygon( _g, shadowPoint );
		//固定点の描画
		_g.setColor( _fillColor );
		SKITPanelHelper.fillPolygon( _g, points );
	}
	
	/**
	 * 選択点を描画する。
	 * @param _g
	 * @param _featurePoint
	 * @param _shadowColor
	 * @param _fillColor 
	 */
	private static void drawSelectedPoint( Graphics2D _g, Point _featurePoint,
		Color _shadowColor, Color _fillColor ) {
		// 影の描画
		Point shadowDrawPoint = _featurePoint.move( SHADOW_DISTANCE, SHADOW_DISTANCE, 0.0 );
		_g.setColor( _shadowColor );
		SKITPanelHelper.fillCircle( _g, shadowDrawPoint, FEATURE_POINT_RADIUS );
		// 選択点の描画
		_g.setColor( _fillColor );
		SKITPanelHelper.fillCircle( _g, _featurePoint, FEATURE_POINT_RADIUS );
	}
}
