package jp.sagalab.skit;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.util.List;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jftk.fragmentation.Fragment;
import jp.sagalab.jftk.fragmentation.IdentificationFragment;
import jp.sagalab.jftk.fragmentation.PartitionFragment;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.curve.ParametricCurve;

/**
 *
 * @author kaneko
 */
public class FragmentLayer extends Component {
	private static final Stroke HOT_FUZZY_POINT_STROKE = new BasicStroke( 0.8f );
	private static final Color HOT_FUZZY_POINT_COLOR_DRAW = new Color( 120, 206, 44, 230 );
	private static final Color HOT_FUZZY_POINT_COLOR = new Color( 120, 206, 44, 230 );
	private static final Stroke ACTIVE_FUZZY_POINT_STROKE = new BasicStroke( 0.8f );
	private static final Color ACTIVE_FUZZY_POINT_COLOR_DRAW = new Color( 101,156, 55, 220 );
	private static final Color ACTIVE_FUZZY_POINT_COLOR = new Color( 101,156, 55, 220 );
	private static final Color PARTITION_FUZZY_POINT_COLOR = new Color( 222, 98, 50, 255 );
	private final ReadOnlySKITModel m_model;

	public static FragmentLayer create( ReadOnlySKITModel _model, int _width, int _height ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		if ( _width <= 0 ) {
			throw new IllegalArgumentException( "_width <= 0" );
		}
		if ( _height <= 0 ) {
			throw new IllegalArgumentException( "_height <= 0" );
		}
		return new FragmentLayer( _model, _width, _height );
	}

	private FragmentLayer( ReadOnlySKITModel _model, int _width, int _height ) {
		m_model = _model;
		setBackground( new Color( 0, 0, 0, 0 ) );
		setSize( _width, _height );
	}

	@Override
	public void paint( Graphics _g ) {
		Graphics2D g = (Graphics2D) _g;

		TransformMatrix modelToView = m_model.getViewToModel().inverse();

		// アクティブな図形群の描画
		drawActiveShapes( g, m_model.getActiveShapes(), modelToView, false );
		// ホットな図形群の描画
		drawActiveShapes( g, m_model.getHotShapes(), modelToView, true );
	}

	/**
	 * アクティブ図形群を描画する。
	 * @param _g
	 * @param _shapes
	 * @param _modelToView
	 * @param _isHot
	 */
	private static void drawActiveShapes( Graphics2D _g, List<Shape> _shapes, TransformMatrix _modelToView, boolean _isHot ) {
		// 図形群描画
		for ( Shape s : _shapes ) {
			// フラグメント列描画
			drawFragments( _g, s.fragments().toArray( new Fragment[ s.fragments().size() ] ), _modelToView, _isHot );
		}
	}

	/**
	 * フラグメント列を描画する。
	 * @param _g
	 * @param _fragments
	 * @param _modelToView
	 * @param _isHot
	 */
	private static void drawFragments( Graphics2D _g, Fragment[] _fragments, TransformMatrix _modelToView, boolean _isHot ) {
		Stroke fuzzyPointStroke = _isHot ? HOT_FUZZY_POINT_STROKE : ACTIVE_FUZZY_POINT_STROKE;

		_g.setStroke( fuzzyPointStroke );
		for ( Fragment fragment : _fragments ) {
			if ( fragment.getClass() == IdentificationFragment.class ) {
				drawIdentificationFragment( _g,
					 fragment.curve().transform( _modelToView ), _isHot );
			} else if ( fragment.getClass() == PartitionFragment.class ) {
				drawPartitionFragment( _g, (PartitionFragment) fragment.transform( _modelToView ) );
			}
		}
	}

	/**
	 * 同定フラグメントを描画する。
	 * @param _g
	 * @param _fragment
	 * @param _isHot
	 */
	private static void drawIdentificationFragment( Graphics2D _g, ParametricCurve _fragment, boolean _isHot ) {
		Color fuzzyPointColor = _isHot ? HOT_FUZZY_POINT_COLOR : ACTIVE_FUZZY_POINT_COLOR;
		Color fuzzyPointColorDraw = _isHot ? HOT_FUZZY_POINT_COLOR_DRAW : ACTIVE_FUZZY_POINT_COLOR_DRAW;

		// フラグメント描画
		_g.setColor( fuzzyPointColor );
		SKITPanelHelper.drawFragment( _g, _fragment );
		// フラグメントの始点と終点を強調
		_g.setColor( fuzzyPointColorDraw );
		Point[] startAndEnd = new Point[]{ _fragment.evaluateAtStart(), _fragment.evaluateAtEnd() };
		SKITPanelHelper.drawFuzzyPoints( _g, startAndEnd.clone() );
	}

	/**
	 * 区切りフラグメントを描画する。
	 * @param _g
	 * @param _partition
	 * @param _modelToView
	 */
	private static void drawPartitionFragment( Graphics2D _g, PartitionFragment _partition ) {
		_g.setColor( PARTITION_FUZZY_POINT_COLOR );
		// 区切り点描画
		Point body = _partition.body();
		SKITPanelHelper.drawCircle( _g, body, body.fuzziness() );
		// 接続の足を描画
		Point start = _partition.start();
		Point end = _partition.end();
		if ( start != null ) {
			SKITPanelHelper.drawLine( _g, start, body );
		}
		if ( end != null ) {
			SKITPanelHelper.drawLine( _g, end, body );
		}
	}
}
