package jp.sagalab.skit;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Timer;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.curve.SplineCurve;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.curve.ParametricEvaluable.EvaluationType;
import org.apache.batik.transcoder.TranscoderException;
import org.apache.batik.transcoder.TranscoderOutput;
import org.w3c.dom.Document;
import org.apache.batik.transcoder.image.PNGTranscoder;
import org.apache.batik.transcoder.TranscoderInput;
import org.apache.batik.transcoder.image.ImageTranscoder;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author kaneko
 */
public class GestureAnimationData extends Component implements ActionListener, AnimationData {
	private static final Color GESTURE_FRAGMENT_COLOR = new Color( 1f, 0f, 0f, 0.2f );
	private static final Stroke GESTURE_FRAGMENT_STROKE = new BasicStroke( 1f );
	private final Timer m_animator;
	private final BufferedImage m_image;
	private final int m_gradation;
	private int m_alpha;

	private static class BufferedImageTranscoder extends ImageTranscoder {
		@Override
		public BufferedImage createImage( int _width, int _height ) {
			BufferedImage img = new BufferedImage( _width, _height, BufferedImage.TYPE_INT_ARGB );
			return img;
		}

		@Override
		public void writeImage( BufferedImage _img, TranscoderOutput _output ) {
			this.m_img = _img;
		}

		public BufferedImage getBufferedImage() {
			return m_img;
		}
		private BufferedImage m_img = null;
	}

	public static GestureAnimationData create( SplineCurve _fsc, Document _svgDoc, BufferedImage _command,
		int _gradation, int _fps ) {
		if ( _fsc == null ) {
			throw new NullPointerException( "_fsc is null." );
		}
		if ( _gradation <= 1 ) {
			throw new IllegalArgumentException( "_gradation <= 1" );
		}
		if ( _fps <= 0 ) {
			throw new IllegalArgumentException( "_fps <= 0" );
		}
		return new GestureAnimationData( _fsc, _svgDoc, _command, _gradation, _fps );
	}

	private GestureAnimationData( SplineCurve _fsc, Document _svgDoc, BufferedImage _command, int _gradation,
		int _fps ) {
		m_gradation = _gradation;
		m_animator = new Timer( (int) Math.round( 1000.0 / _fps ), this );
		m_alpha = 0;

		Rectangle repaintArea = calcRepaintArea( _fsc, _command );
		// コンポーネントの原点への位置合わせのための変換行列
		TransformMatrix mat = TransformMatrix.translation( -repaintArea.x, -repaintArea.y, 0.0 );

		// 描画画像の作成
		m_image = createImage( _fsc.transform( mat ), _svgDoc, _command, repaintArea );

		setBackground( new Color( 0, 0, 0, 0 ) );
		setLocation( repaintArea.x, repaintArea.y );
		setSize( repaintArea.width, repaintArea.height  );
	}

	@Override
	public void paint( Graphics _g ) {
		Graphics2D g = (Graphics2D) _g;

		// アルファ値を取得
		float alpha = getAlpha() / (float) ( m_gradation - 1 );
		// 透明度を設定
		g.setComposite( AlphaComposite.getInstance( AlphaComposite.SRC_OVER, alpha ) );

		g.drawImage( m_image, 0, 0, null );
	}

	private static BufferedImage createImage( SplineCurve _fsc, Document _svgDoc, BufferedImage _command,
		Rectangle _repaintArea ) {
		BufferedImage image = ImageCreater.createTransparentImage( _repaintArea.width, _repaintArea.height );
		Graphics2D g = image.createGraphics();
		int svgWidth = 0;
		int svgHeight = 0;

		//xml文書から画像サイズ(width,height)のデータを取得
		if ( _svgDoc != null ) {
			String width = pickOutNamedNodeValue( _svgDoc, "width" );
			String height = pickOutNamedNodeValue( _svgDoc, "height" );
			if ( width != null ) {
				svgWidth = Integer.parseInt( width.replaceAll( "[^0-9]", "" ) );
			}
			if ( height != null ) {
				svgHeight = Integer.parseInt( height.replaceAll( "[^0-9]", "" ) );
			}
		}

		Rectangle gestureFragmentArea = calcGestureFragmentArea( _fsc );
		java.awt.Point imagePosition = calcImagePosition( gestureFragmentArea.width, gestureFragmentArea.height );
		double scalingRatio = calcImageScalingRatio( svgWidth, svgHeight, gestureFragmentArea.width, gestureFragmentArea.height );

		// フラグメントを描画
		g.setStroke( GESTURE_FRAGMENT_STROKE );
		g.setColor( GESTURE_FRAGMENT_COLOR );
		SKITPanelHelper.drawFragment( g, _fsc );

		// ジェスチャのシンボル画像を描画
		int symbolHeight = 0;
		if ( _svgDoc != null ) {
			// 描画画像サイズの決定
			int symbolWidth = (int) Math.round( svgWidth * scalingRatio );
			symbolHeight = (int) Math.round( svgHeight * scalingRatio );

			// 拡大後の画像イメージを作成し、描画する
			BufferedImageTranscoder imageTranscoder = new BufferedImageTranscoder();
			TranscoderInput input = new TranscoderInput( _svgDoc );
			imageTranscoder.addTranscodingHint(PNGTranscoder.KEY_WIDTH, (float) symbolWidth);
			imageTranscoder.addTranscodingHint(PNGTranscoder.KEY_HEIGHT, (float) symbolHeight);
			try {
				imageTranscoder.transcode( input, null );
			} catch ( TranscoderException ex ) {
				Logger.getLogger( GestureAnimationData.class.getName() ).log( Level.SEVERE, null, ex );
			}
			BufferedImage svgImage = imageTranscoder.getBufferedImage();
			g.drawImage( svgImage, imagePosition.x, imagePosition.y, symbolWidth, symbolHeight, null );
		}

		// ジェスチャのコマンド名の画像を描画
		int commandWidth = _command.getWidth();
		int commandHeight = _command.getHeight();
		g.drawImage( _command,imagePosition.x, imagePosition.y + symbolHeight, commandWidth, commandHeight, null );

		g.dispose();

		return image;
	}

	/**
	 * ジェスチャ画像の表示範囲の計算。
	 * @param _fsc
	 * @return
	 */
	private static Rectangle calcRepaintArea( SplineCurve _fsc, BufferedImage _command ) {
		// ジェスチャフラグメントの範囲
		Rectangle rectangle = calcGestureFragmentArea( _fsc );
		// コマンド画像の幅と高さを考えて計算
		return new Rectangle( rectangle.x, rectangle.y,
			Math.max( rectangle.width , _command.getWidth() ), rectangle.height + _command.getHeight() );
	}

	/**
	 * ジェスチャフラグメントの範囲の計算。
	 * @param _fsc
	 * @return
	 */
	private static Rectangle calcGestureFragmentArea( SplineCurve _fsc ) {
		double minX = Double.POSITIVE_INFINITY;
		double maxX = Double.NEGATIVE_INFINITY;
		double minY = Double.POSITIVE_INFINITY;
		double maxY = Double.NEGATIVE_INFINITY;
		int num = Math.max( (int) Math.ceil( _fsc.range().length() / 0.01 ), 2 );
		Point[] points = _fsc.evaluateAll( num, EvaluationType.TIME );
		for ( Point p : points ) {
			double fuzziness = p.fuzziness();
			minX = Math.min( p.x() - fuzziness, minX );
			maxX = Math.max( p.x() + fuzziness, maxX );
			minY = Math.min( p.y() - fuzziness, minY );
			maxY = Math.max( p.y() + fuzziness, maxY );
		}
		return new Rectangle( (int) Math.floor( minX ), (int) Math.floor( minY ),
			(int) Math.ceil( maxX - minX ), (int) Math.ceil( maxY - minY ) );
	}

	/**
	 * ジェスチャ表示位置の計算。
	 *
	 * @param _fsc 曲線。
	 */
	private static java.awt.Point calcImagePosition( int _width, int _height ) {
		int size = Math.min( _width, _height );
		int x = ( _width - size ) / 2;
		int y = ( _height - size ) / 2;

		return new java.awt.Point( x, y );
	}

	/**
	 * ジェスチャ表示倍率の計算
	 *
	 * @param _curve 曲線。
	 */
	private static double calcImageScalingRatio( int _svgWidth, int _svgHeight, int _width, int _height ) {
		int imageHeight = _svgHeight;
		int size = Math.min( _width, _height );
		double widthScalingRatio = size / (double) _svgWidth;
		double heightScalingRatio = size / (double) imageHeight;

		return Math.min( widthScalingRatio, heightScalingRatio );
	}

	/**
	 * xml文書のノードから指定した属性値の取得
	 *
	 * @param String 取得する属性の名前
	 */
	private static String pickOutNamedNodeValue( Document _svgDoc, String _named ) {
		String nodeValue = null;
		NodeList nodes = _svgDoc.getChildNodes();
		int nodesLength = nodes.getLength();
		for ( int i = 0; i < nodesLength; i++ ) {
			NamedNodeMap attrs = nodes.item( i ).getAttributes();
			if ( attrs != null ) {
				Node namedItem = attrs.getNamedItem( _named );
				if ( namedItem != null ) {
					nodeValue = namedItem.getNodeValue();
				}
			}
		}
		return nodeValue;
	}

	@Override
	public void startAnimation() {
		m_animator.start();
	}

	private int getAlpha() {
		int alpha = m_alpha;
		int gradation = m_gradation - 1;

		return Math.min( alpha, gradation ) - Math.max( alpha - gradation, 0 );
	}

	@Override
	public void actionPerformed( ActionEvent ae ) {
		if ( updateScene() ) {
			repaint();
		} else {
			m_animator.stop();
			getParent().remove( this );
		}
	}

	@Override
	public boolean updateScene() {
		if ( m_alpha < 2 * ( m_gradation - 1 ) ) {
			++m_alpha;
			return true;
		} else {
			return false;
		}
	}
}
