package jp.sagalab.skit;

import java.awt.event.MouseEvent;
import java.util.LinkedList;
import javax.swing.event.MouseInputAdapter;
import jp.sagalab.jftk.Point;

/**
 * 入力点列のコントローラ。
 */
public class InputPointsController extends MouseInputAdapter {
	/** ピッキング入力時間しきい値[sec] */
	private static final double PICKING_TIME = 0.17;
	/** ピッキング入力距離しきい値[inch] */
	private static final double PICKING_INCH_DISTANCE = 0.3 / 2.54;
	/** 1インチあたりのピクセル数(pixel per inch) */
	private final double m_ppi;
	/** 入力点列 */
	private final LinkedList<Point> m_points;
	/** モデル */
	private final InputOnlySKITModel m_model;
	/** 入力点列が同じ時刻を持つ現象を阻止 */
	private double m_lastInputTime = 0.0;

	public static InputPointsController create( InputOnlySKITModel _model, double _ppi ) {
		return new InputPointsController( _model, new LinkedList<Point>(), _ppi );
	}

	protected InputPointsController( InputOnlySKITModel _model, LinkedList<Point> _points, double _ppi ) {
		m_model = _model;
		m_points = _points;
		m_ppi = _ppi;
	}

	@Override
	public void mousePressed( MouseEvent me ) {
		// マウスの第1ボタンのみ押している時
		if( me.getModifiers()== MouseEvent.BUTTON1_MASK ) {
			Point point = Point.createXYZT( me.getX(), me.getY(), 0.0, System.currentTimeMillis() * 0.001 );
			// 入力点列をクリア
			m_points.clear();

			// 学習停止
			m_model.cancelLearning();

			// 点を追加
			m_points.add( point );

			m_lastInputTime = point.time();
		}
	}

	@Override
	public void mouseDragged( MouseEvent me ) {
		// マウスの第1ボタンのみ押している時
		if ( me.getModifiers() == MouseEvent.BUTTON1_MASK ) {
			Point point = Point.createXYZT( me.getX(), me.getY(), 0.0, System.currentTimeMillis() * 0.001 );
			if ( point.time() != m_lastInputTime ) {
				// 点を追加
				m_points.add( point );

				m_lastInputTime = point.time();
			}
		}
	}

	@Override
	public void mouseReleased( MouseEvent me ) {		
		// マウスの第1ボタンのみ押している時
		if( me.getModifiers()== MouseEvent.BUTTON1_MASK ) {
			// 点を追加
			Point point = Point.createXYZT( me.getX(), me.getY(), 0.0, System.currentTimeMillis() * 0.001 );
			if ( point.time() != m_lastInputTime ) {
				// 点を追加
				m_points.add( point );

				m_lastInputTime = point.time();
			}

			Point[] points = m_points.toArray( new Point[ m_points.size() ] );
			if ( isPicking( points ) ) {
				// 1ピクセルあたりのインチ(inch per pixel)を算出
				double inchPerPixel = 1.0 / m_ppi;
				m_model.picking( point, inchToCm( inchPerPixel ), inchToCm( PICKING_INCH_DISTANCE ) );
			} else {
				// モデルに入力点列を渡す
				m_model.putPoints( points );
			}
		}
	}

	/**
	 * 入力点列がピッキング操作に該当するかどうかを判定する。
	 * @param _inputs 入力点列
	 * @return ピッキング操作に該当する場合はtrue。
	 */
	private boolean isPicking( Point[] _inputs ) {
		if ( _inputs.length == 0 ) {
			throw new IllegalArgumentException( "_inputs.length == 0" );
		}
		// 入力開始点
		Point start = _inputs[0];
		Point end = _inputs[_inputs.length - 1];

		// 入力時間がしきい値未満かどうか
		boolean isPickingTime = end.time() - start.time() < PICKING_TIME;

		// 入力点列の距離がしきい値未満かどうか
		boolean isPickingDistance = pixelToInch( Point.length( _inputs ) ) < PICKING_INCH_DISTANCE;

		// 時間・距離ともに条件を満たしていればピッキング操作
		return isPickingTime && isPickingDistance;
	}

	/**
	 * 指定されたinch単位の距離ををcm距離に変換して返す。
	 * @param _inch 変換元となるinch単位の距離。
	 * @return 変換後のcm単位の距離。
	 */
	private static double inchToCm( double _inch ) {
		return _inch * 2.54;
	}

	/**
	 * pixel単位の距離をinch距離に変換して返す。
	 * @param _pixel 変換元となるpixel単位の距離。
	 * @return 変換後のinch単位の距離。
	 */
	private double pixelToInch( double _pixel ) {
		// 1インチあたりのピクセル数(pixel per inch)を算出
		return _pixel / m_ppi;
	}
	
}
