package jp.sagalab.skit;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.InputEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JFrame;
import javax.swing.event.MouseInputListener;

/**
 * キーコントローラ。
 * @author kaneko
 */
public class KeyController extends KeyAdapter {
	/** モデル */
	private final SKITModel m_model;
	/** XMLサーバ */
	private final XMLServer m_server;
	/** コンテナ */
	private final Container m_container;
	/** 相対図形空間操作のコントローラ */
	private final RelativeModelSpaceController m_relativeModelSpaceController;
	/** 図形描画モードフラグ */
	private boolean m_isInputPointsMode = true;
	/** 1インチあたりのピクセル数(pixel per inch) */
	private final double m_ppi;
	/** 回転量の基準値 */
	private static final double BASE_OF_ROTATE = Math.PI / 12;
	/** 拡大・縮小率の基準値 */
	private static final double BASE_OF_SCALE = 2.0;

	public static KeyController create( SKITModel _model, XMLServer _server, Container _container, RelativeModelSpaceController _relative, double _ppi ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		if ( _server == null ) {
			throw new NullPointerException( "_server is null." );
		}
		if ( _container == null ) {
			throw new NullPointerException( "_container is null." );
		}
		if ( _relative == null ) {
			throw new NullPointerException( "_relative is null." );
		}
		if ( _ppi <= 0 ) {
			throw new IllegalArgumentException( "_ppi <= 0" );
		}
		if ( Double.isNaN( _ppi ) ) {
			throw new IllegalArgumentException( "_ppi is NaN." );
		}
		if ( Double.isInfinite( _ppi ) ) {
			throw new IllegalArgumentException( "_ppi is inf." );
		}
		return new KeyController( _model, _server, _container, _relative, _ppi );
	}

	private KeyController( SKITModel _model, XMLServer _server, Container _container, RelativeModelSpaceController _relative, double _ppi ) {
		m_model = _model;
		m_server = _server;
		m_container = _container;
		m_relativeModelSpaceController = _relative;
		m_ppi = _ppi;
	}

	@Override
	public void keyPressed( KeyEvent _ev ) {
		double transfom = m_model.getGridSpace().grid().scalalize();
		switch ( _ev.getKeyCode() | _ev.getModifiersEx() ) {
			case KeyEvent.VK_Q | InputEvent.CTRL_DOWN_MASK:
				// 親コンテナを不可視にして，親コンテナ上の全てのコンポーネントを削除
				closeFrame();
				// 空のXMLデータを送信
				m_server.send();
				try {
					m_server.close();
				} catch ( IOException ex ) {
					Logger.getLogger( KeyController.class.getName() ).log( Level.SEVERE, null, ex );
				} catch ( IllegalStateException ex ) {
					Logger.getLogger( KeyController.class.getName() ).log( Level.SEVERE, null, ex );
				}
				break;
			case KeyEvent.VK_ESCAPE:
				if ( m_model.isChangedViewToModelMatrix() ) {
					ModelSpaceInitializer initializer = ModelSpaceInitializer.create( m_model, m_container );
					initializer.startAnimation();
				}
				break;
			case KeyEvent.VK_ENTER:
				// 親コンテナを不可視にして，親コンテナ上の全てのコンポーネントを削除
				closeFrame();
				// 図形群のXMLデータを送信
				try {
					m_server.send( m_model.getShapes() );
					m_server.close();
				} catch ( IOException ex ) {
					Logger.getLogger( KeyController.class.getName() ).log( Level.SEVERE, null, ex );
				} catch ( IllegalStateException ex ) {
					Logger.getLogger( KeyController.class.getName() ).log( Level.SEVERE, null, ex );
				}
				System.out.println( "exit" );
				break;
			case KeyEvent.VK_U:
				m_model.undo();
				break;
			case KeyEvent.VK_R:
				m_model.redo();
				break;
			case KeyEvent.VK_SPACE:
				Dimension dimension = m_container.getPreferredSize();
				if ( m_isInputPointsMode ) {
					// 絶対図形空間操作モードに切り替え
					// 絶対図形空間操作コントローラをマウスリスナーに登録。
					MouseAbsoluteModelSpaceController absoluteModelSpaceController;
					try {
						absoluteModelSpaceController = MouseAbsoluteModelSpaceController.create( m_model );
					} catch ( IllegalStateException ex ) {
						return;
					}
					switchMouseInputListener( absoluteModelSpaceController );
					// 特徴点列レイヤーをコンポーネントに登録
					FeaturePointsLayer featurePointsLayer = FeaturePointsLayer.create( m_model, dimension.width, dimension.height );
					setLayer( featurePointsLayer, 0 );
				} else {
					// 図形描画モードに切り替え
					// スクラッチレイヤーをコンポーネントに登録
					ScratchLayer scratchLayer = ScratchLayer.create( dimension.width, dimension.height );
					setLayer( scratchLayer, 0 );
					// 入力点列のコントローラをマウスリスナーに登録
					InputPointsController inputPointsController = InputPointsController.create( m_model, m_ppi );
					switchMouseInputListener( inputPointsController );
					// スクラッチレイヤーをマウスリスナーに登録
					m_container.addMouseListener( scratchLayer );
					m_container.addMouseMotionListener( scratchLayer );
					m_model.resetFeaturePoints();
				}
				m_container.repaint();
				m_isInputPointsMode = !m_isInputPointsMode;
				break;
				// ----- ここから下は、相対図形空間操作についての処理 -----
			// TODO 相対図形空間操作の移動量、回転量、拡大・縮小率は要検討
			// 上への移動
			case KeyEvent.VK_W:
			case KeyEvent.VK_UP:
				m_relativeModelSpaceController.move( 0.0, transfom, 0.0 );
				break;
			// 下への移動
			case KeyEvent.VK_S:
			case KeyEvent.VK_DOWN:
				m_relativeModelSpaceController.move( 0.0, -transfom, 0.0 );
				break;
			// 左への移動
			case KeyEvent.VK_A:
			case KeyEvent.VK_LEFT:
				m_relativeModelSpaceController.move( transfom, 0.0, 0.0 );
				break;
			// 右への移動
			case KeyEvent.VK_D:
			case KeyEvent.VK_RIGHT:
				m_relativeModelSpaceController.move( -transfom, 0.0, 0.0 );
				break;
			// 時計周りに回転 (15度刻み)
			case KeyEvent.VK_E:
			case KeyEvent.VK_D | InputEvent.CTRL_DOWN_MASK:
			case KeyEvent.VK_RIGHT | InputEvent.CTRL_DOWN_MASK:
				m_relativeModelSpaceController.rotate( -BASE_OF_ROTATE );
				break;
			// 反時計周りに回転 (15度刻み)
			case KeyEvent.VK_Q:
			case KeyEvent.VK_A | InputEvent.CTRL_DOWN_MASK:
			case KeyEvent.VK_LEFT | InputEvent.CTRL_DOWN_MASK:
				m_relativeModelSpaceController.rotate( BASE_OF_ROTATE );
				break;
			// モデル空間の拡大
			case KeyEvent.VK_W | InputEvent.CTRL_DOWN_MASK:
			case KeyEvent.VK_UP | InputEvent.CTRL_DOWN_MASK:
				m_relativeModelSpaceController.scale( 1.0 / BASE_OF_SCALE );
				break;
			// モデル空間の縮小
			case KeyEvent.VK_S | InputEvent.CTRL_DOWN_MASK:
			case KeyEvent.VK_DOWN | InputEvent.CTRL_DOWN_MASK:
				m_relativeModelSpaceController.scale( BASE_OF_SCALE );
				break;
			// ----- ここまで -----
			default:
		}
	}

	/**
	 * MouseInputリスナーの切り替え。
	 * @param _listener
	 * @param _layer
	 */
	private void switchMouseInputListener( MouseInputListener _listener ) {
		MouseListener[] mouseListeners = m_container.getMouseListeners();
		MouseMotionListener[] mouseMotionListeners = m_container.getMouseMotionListeners();
		for ( MouseListener mouseListener : mouseListeners ) {
			m_container.removeMouseListener( mouseListener );
		}
		for ( MouseMotionListener mouseMotionListener : mouseMotionListeners ) {
			m_container.removeMouseMotionListener( mouseMotionListener );
		}

		m_container.addMouseListener( _listener );
		m_container.addMouseMotionListener( _listener );
	}

	/**
	 * レイヤーの設定。
	 * @param _layer レイヤー。
	 * @param _index コンポーネントのインデックス。
	 */
	private void setLayer( Component _layer, int _index ) {
		m_container.remove( _index );
		m_container.add( _layer, _index );
	}

	private JFrame getParentJFrame() {
		Component component = m_container.getParent();
		while ( !( component instanceof JFrame ) ) {
			component = component.getParent();
		}
		return (JFrame) component;
	}

	private void closeFrame() {
		JFrame frame = getParentJFrame();
		frame.setVisible( false );
		frame.getContentPane().removeAll();
	}
}
