package jp.sagalab.skit;

import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.TooManyListenersException;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

/**
 * Magellanとの通信を実現するクラス。
 * @author fujinaga
 */
public final class Magellan implements SerialPortEventListener, DeviceInterface {
	/** Magellanへの出力ストリーム */
	private OutputStream m_outputStream = null;
	/** シリアルポート */
	private final SerialPort m_serialPort;
	/** 入力ストリーム */
	private final InputStream m_inputStream;
	/** 処理中のバッファ */
	//private ByteBuffer localBuffer;
	/** ロガー */
	private static final Logger m_logger = Logger.getLogger( Magellan.class.getName() );
	/** シリアルポートイベントを受信するまでの受信バイト数 */
	private static final int M_RECEIVE_THRESHOLD = 256;
	/** Magellanのずらし値 */
	private float[] m_position = new float[ 3 ];
	/** Magellanのひねり値 */
	private float[] m_rotation = new float[ 3 ];
	/** ボタンの数 */
	private final int M_BUTTONS_NUM = 9;
	/** ボタン押下状態（押下時にtrue） */
	private boolean[] m_buttonsState = new boolean[ M_BUTTONS_NUM ];
	/** テンポラリバッファの大きさ */
	private static final int M_TEMP_BUFFER_SIZE = 32;
	/** データの断片を一時的に保存しておくためのテンポラリバッファ */
	private final byte[] m_tempBuffer = new byte[ M_TEMP_BUFFER_SIZE ];
	/** ずらし、ひねりのデータのヘッダ */
	private static final byte M_HEADER_POS_EVENT = 'd' & 0xff;
	/** キーイベントのデータのヘッダ */
	private static final byte M_HEADER_KEY_EVENT = 'k' & 0xff;
	/** 改行文字 */
	private static final byte M_LINEFEED_CODE = '\r' & 0xff;

	static Magellan setting( String _portName ) throws TooManyListenersException, IOException, NoSuchPortException, PortInUseException, UnsupportedCommOperationException {
		return new Magellan( _portName );
	}

	/**
	 * ポート名を指定してインスタンス化する。
	 * @param _portName Magellanが使用しているポート名
	 * @throws java.util.TooManyListenersException シリアルポートのイベントリスナが規定数を越えたとき
	 * @throws java.io.IOException 入出力でエラーが発生したとき
	 * @throws gnu.io.NoSuchPortException 指定したポートが見つからないとき
	 * @throws gnu.io.PortInUseException 指定したポートが既に使用されているとき
	 * @throws gnu.io.UnsupportedCommOperationException サポートされていない処理が呼び出されたとき
	 */
	private Magellan( String _portName ) throws TooManyListenersException, IOException, NoSuchPortException, PortInUseException, UnsupportedCommOperationException {
		// ロガーの設定
		Handler handler = new ConsoleHandler();
		handler.setFormatter( new SimpleFormatter() );
		handler.setLevel( Level.INFO );
		m_logger.addHandler( handler );

		Handler fileHandler = new FileHandler( Magellan.class.getName() + ".log" );
		fileHandler.setFormatter( new SimpleFormatter() );
		fileHandler.setLevel( Level.FINEST );
		m_logger.addHandler( fileHandler );
		m_logger.setLevel( Level.FINEST );

		// ポートを開く
		CommPortIdentifier portID = CommPortIdentifier.getPortIdentifier( _portName );
		m_serialPort = (SerialPort) portID.open( SKITPanel.class.getName(), 3000 );
		m_serialPort.setSerialPortParams( 9600, SerialPort.DATABITS_8, SerialPort.STOPBITS_2, SerialPort.PARITY_NONE );
		// 入出力先ストリームの取得
		m_outputStream = m_serialPort.getOutputStream();
		m_inputStream = m_serialPort.getInputStream();
		// 入出力でハンドシェイクを行うように設定
		m_serialPort.setFlowControlMode( SerialPort.FLOWCONTROL_RTSCTS_IN | SerialPort.FLOWCONTROL_RTSCTS_OUT );

		// イベントリスナの設定
		m_serialPort.notifyOnDataAvailable( true );
		m_serialPort.enableReceiveThreshold( M_RECEIVE_THRESHOLD );
		m_serialPort.enableReceiveTimeout( 200 );
		m_serialPort.addEventListener( this );
		// 初期設定を行う
		initialSettings();
	}

	/**
	 * Magellanの初期設定を行う。
	 * @throws java.io.IOException 入出力エラーが発生したとき
	 */
	private void initialSettings() throws IOException {
		// コマンドを用意
		final byte[] shortBeepCommand = { 'b' & 0xff, (byte) 0x08, M_LINEFEED_CODE };
		final byte[] setDataRate60To80Command = { 'p' & 0xff, (byte) 0x04, (byte) 0x02, M_LINEFEED_CODE };
		final byte[] setSensivityCommand = { 'q' & 0xff, (byte) 0x00, (byte) 0x00, M_LINEFEED_CODE };
		final byte[] setNullRadiusCommand = { 'n' & 0xff, (byte) 0x00, M_LINEFEED_CODE };
		final byte[] zeroingCommand = { 'z' & 0xff, M_LINEFEED_CODE };
		final byte[] lightCommand = { 'l' & 0xff, (byte) 0x07, M_LINEFEED_CODE };
		final byte[] setOperationModeCommand = { 'm' & 0xff, (byte) 0x07, M_LINEFEED_CODE };
		try {
			// Magellanにコマンドを送信
			Thread.sleep( 3000 );
		} catch ( InterruptedException ex ) {
			Logger.getLogger( Magellan.class.getName() ).log( Level.SEVERE, null, ex );
		}
		m_outputStream.write( setDataRate60To80Command );
		m_outputStream.write( setNullRadiusCommand );
		m_outputStream.write( zeroingCommand );
		m_outputStream.write( setSensivityCommand );
		m_outputStream.write( setOperationModeCommand );
		m_outputStream.write( lightCommand );
		m_outputStream.write( shortBeepCommand );
		m_outputStream.flush();
	}

	/**
	 * Magellanが使用しているシリアル通信ポートと入出力ストリームを閉じる。
	 * @throws java.io.IOException 入出力エラーが発生したとき
	 */
	public void closePort() throws IOException {
		/*if( localBuffer.asReadOnlyBuffer().get() != 0 ){
		localBuffer.clear();
		}*/
		m_outputStream.close();
		m_inputStream.close();
		m_serialPort.notifyOnDataAvailable( false );
		m_serialPort.close();
	}

	@Override
	public void update() {
	}

	@Override
	public int getButtonsNum() {
		return M_BUTTONS_NUM;
	}

	@Override
	public boolean[] getButtonsState() {
		return m_buttonsState.clone();
	}

	@Override
	public float[] getPosition() {
		return m_position.clone();
	}

	@Override
	public float[] getOrientation() {
		return m_rotation.clone();
	}

	/*public void deleteRemainedBuffer(){
	localBuffer.clear();
	}*/
	@Override
	public void serialEvent( SerialPortEvent _ev ) {
		if ( _ev.getEventType() == SerialPortEvent.DATA_AVAILABLE ) {
			byte[] buf = new byte[ M_RECEIVE_THRESHOLD ];
			try {
				int numOfRead = m_inputStream.read( buf );
				parse( buf, numOfRead );
			} catch ( IOException ex ) {
				m_logger.log( Level.SEVERE, null, ex );
			}
		}
	}

	/**
	 * 受信したデータを解析する。
	 * @param _buf 受信したデータ
	 * @param _size データの数（バイト）
	 */
	private void parse( byte[] _buf, int _size ) {
		// ローカルバッファを生成し、一時バッファに受信データを連結させたものを格納する
		ByteBuffer localBuffer = ByteBuffer.allocate( M_TEMP_BUFFER_SIZE + _size );
		localBuffer.put( removeZeros( m_tempBuffer ) );
		localBuffer.put( _buf, 0, _size );
		localBuffer.position( 0 );

		// バッファ内の改行文字がなくなるまで解析する
		int dataHead = -1;
		while ( true ) {
			// 改行文字を探索し、その次のバイト要素の内容で分岐
			dataHead = getIndex( localBuffer, M_LINEFEED_CODE );
			if ( dataHead == -1 ) {
				break;
			}
			localBuffer.position( dataHead + 1 );
			switch ( localBuffer.get() ) {
				case M_HEADER_POS_EVENT:
					if ( getIndex( localBuffer, M_LINEFEED_CODE ) != -1 ) {
						byte[] dataSet = new byte[ 24 ]; // X,Y,Z,A,B,C各4Byteずつ
						localBuffer.get( dataSet );
						receivedPositionEvent( dataSet );
					}
					break;
				case M_HEADER_KEY_EVENT:
					if ( getIndex( localBuffer, M_LINEFEED_CODE ) != -1 ) {
						byte[] dataSet = new byte[ 3 ];
						localBuffer.get( dataSet );
						receivedKeyEvent( dataSet );
					}
					break;
				case M_LINEFEED_CODE:
					localBuffer.position( localBuffer.position() - 1 );
					break;
				default:
					break;
			}
		}

		//try {
		// 残ったものを一時バッファに保管
		if ( localBuffer.asReadOnlyBuffer().get() != 0 ) {
			localBuffer.get( m_tempBuffer, 0, localBuffer.remaining() );
		}
		/*} catch( IndexOutOfBoundsException e ){
		//System.out.println("終了時に残る一時バッファを無視します．");
		}*/
	}

	/**
	 * ByteBufferの中から指定した値の要素を探索し、最初に見つかった位置を返す。
	 * ただし、見つからなかった場合は-1を返す
	 * @param _buf 探索を行うバッファ
	 * @param _target 探索対象の数値
	 * @return 探索対象が見つかった場所のインデックス
	 */
	private int getIndex( ByteBuffer _buf, byte _target ) {
		ByteBuffer buffer = _buf.asReadOnlyBuffer();
		while ( buffer.hasRemaining() ) {
			if ( buffer.get() == _target ) {
				return buffer.position() - 1;
			}
		}
		return -1;
	}

	/**
	 * byte[]配列から値が0の要素を取り除いて返す。
	 * @param _buf 元配列
	 * @return 元配列から値が0の要素が取り除かれた配列
	 */
	private byte[] removeZeros( byte[] _buf ) {
		byte[] copiedArray = _buf.clone();
		// 0以外の要素の数を数えて配列を確保
		int zeroCounter = 0;
		for ( byte b : copiedArray ) {
			if ( b != 0 ) {
				++zeroCounter;
			}
		}
		byte[] byteArray = new byte[ zeroCounter ];
		// 0以外の要素を代入して返す
		int i = 0;
		for ( byte b : copiedArray ) {
			if ( b != 0 ) {
				byteArray[i++] = b;
			}
		}
		return byteArray;
	}

	/**
	 * Magellanのボタン押下イベントに対する処理を行う。
	 * @param _buf ヘッダを含まない一件分の生データ
	 */
	private void receivedKeyEvent( byte[] _buf ) {
		byte[] copiedArray = _buf.clone();
		boolean[] buttonsState = new boolean[ M_BUTTONS_NUM ];
		buttonsState[0] = ( copiedArray[2] & 0x01 ) != 0;
		buttonsState[1] = ( copiedArray[0] & 0x01 ) != 0;
		buttonsState[2] = ( copiedArray[0] & 0x02 ) != 0;
		buttonsState[3] = ( copiedArray[0] & 0x04 ) != 0;
		buttonsState[4] = ( copiedArray[0] & 0x08 ) != 0;
		buttonsState[5] = ( copiedArray[1] & 0x01 ) != 0;
		buttonsState[6] = ( copiedArray[1] & 0x02 ) != 0;
		buttonsState[7] = ( copiedArray[1] & 0x04 ) != 0;
		buttonsState[8] = ( copiedArray[1] & 0x08 ) != 0;

		m_buttonsState = buttonsState;
	}

	/**
	 * Magellanに対するずらし、ひねりのイベントに対する処理を行う。
	 * @param _bytes ヘッダを含まない一件分の生データ
	 */
	private void receivedPositionEvent( byte[] _bytes ) {
		int[] pos = new int[ 6 ];
		for ( int j = 0; j < 6; ++j ) {
			byte[] aData = new byte[ 4 ];
			for ( int i = 0; i < 4; ++i ) {
				aData[i] = _bytes[j * 4 + i];
			}
			pos[j] = decode( aData );
		}
		m_position = new float[]{ pos[0], pos[1], pos[2] };
		m_rotation = new float[]{ pos[3], pos[4], pos[5] };
	}

	/**
	 * Magellanから受け取ったバイトデータを数値にデコードする。
	 * @param _byteData 4バイトの受信データが格納された配列
	 * @return 変換して得られた数値
	 */
	private int decode( byte[] _byteData ) {
		return ( ( 0x0f & _byteData[0] ) << 12 | ( 0x0f & _byteData[1] ) << 8 | ( 0x0f & _byteData[2] ) << 4 | ( 0x0f & _byteData[3] ) ) - 0x8000;
	}

}
