package jp.sagalab.skit;

import java.awt.AWTException;
import java.awt.Image;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.SystemTray;
import java.awt.Toolkit;
import java.awt.TrayIcon;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;

public class Main{
	/** SKITが通信に使用するポート番号 */
	private static final int SKIT_PORT_NUM = 4910;
	/** 1インチあたりのピクセル数(pixel per inch) */
	private static final double PPI = Toolkit.getDefaultToolkit().getScreenResolution();
	/** グリッド間隔[cm] */
	private static final double GRID_SIZE = 1.0;

	/**
	 * @param args the command line arguments
	 */
	public static void main( String[] args ) {
		SKITLogger.configure();
		// SKITウィンドウ生成
		final JFrame frame = new JFrame( "SKIT" );
		SwingUtilities.invokeLater( new Runnable(){
			@Override
			public void run() {
				frame.setUndecorated( true );
				frame.setAlwaysOnTop( true );
				frame.setFocusable( false );
			}
		} );

		// サーバソケットを生成
		ServerSocket server = null;
		try {
			server = new ServerSocket( SKIT_PORT_NUM );
		} catch ( IOException ex ) {
			// サーバソケットの生成に失敗したらプログラム終了
			Logger.getLogger( Main.class.getName() ).log( Level.SEVERE, null, ex );
			System.exit( 1 );
		}

		// タスクトレイ生成
		createTaskTray();

		// 接続待ち受けループ
		while ( true ) {
			try {
				// クライアントからの接続を待つ
				Socket socket = server.accept();

				// クライアントからビューポート情報とモデル空間情報を受信
				final XMLServer xmlServer = new XMLServer( socket );
				final ViewportInfo vi = new ViewportInfo();
				final ModelSpaceInfo mi = new ModelSpaceInfo();
				xmlServer.receive( vi, mi );
				SwingUtilities.invokeLater( new Runnable(){
					@Override
					public void run() {
						try {
							// パネルを生成してフレームに追加
							SKITPanel panel = SKITPanel.create( PPI, xmlServer, mi, vi );
							panel.setFocusable( true );
							panel.requestFocusInWindow();

							frame.add( panel );
							frame.setBounds( vi.X(), vi.Y(), vi.width(), vi.height() );
							frame.pack();
							frame.setVisible( true );
						} catch ( AWTException ex ) {
							// スクリーンキャプチャに失敗した場合は再度待ち受け状態にする
							frame.setVisible( false );
							frame.removeAll();
							Logger.getLogger( Main.class.getName() ).log( Level.SEVERE, null, ex );
						}
					}
				} );

			} catch ( IOException ex ) {
				// 再度待ち受け状態にする
				Logger.getLogger( Main.class.getName() ).log( Level.SEVERE, null, ex );
			}
		}
	}

	/**
	 * タスクトレイの生成
	 */
	private static void createTaskTray() {
		SwingUtilities.invokeLater( new Runnable(){
			public void run() {
				// システムトレイを取得
				SystemTray tray = SystemTray.getSystemTray();
				PopupMenu popup = new PopupMenu();

				// マウスモジュールを接続するメニュー項目の生成
				MenuItem openMenu = new MenuItem( "OPEN" );
				openMenu.addActionListener( new ActionListener(){
					@Override
					public void actionPerformed( ActionEvent _event ) {
						Socket socket = null;
						try {
							// SKIT待ち受けポートに接続し、ビューポート情報とモデル空間情報を送信する
							socket = new Socket( "localhost", SKIT_PORT_NUM );
							DummyClient client = new DummyClient( socket );
							// ppiからグリッド間隔(cm)のピクセル数を算出
							double gridSize = ( PPI * GRID_SIZE ) / 2.54;
							client.sendDummyInfo( gridSize );
							new Thread( client ).start();
						} catch ( IOException ex ) {
							Logger.getLogger( Main.class.getName() ).log( Level.SEVERE, null, ex );
							if ( socket != null ) {
								try {
									socket.close();
								} catch ( IOException ignore ) {
								}
							}
						}
					}
				} );

				// プログラムを終了するメニュー項目の生成
				MenuItem exitMenu = new MenuItem( "EXIT" );
				exitMenu.addActionListener( new ActionListener(){
					@Override
					public void actionPerformed( ActionEvent _event ) {
						System.exit( 0 );
					}
				} );

				// ポップアップメニューにメニュー項目を追加
				popup.add( openMenu );
				popup.add( exitMenu );
				// OS名取得
				String osName = System.getProperty( "os.name" ).split( " " )[0];
				Image image=null;
				// システムトレイにアイコンを追加
				switch ( OS.valueOf( osName ) ) {
					case Mac:// Mac用
						String currentPath = new File( "." ).getAbsoluteFile().getParent();
						image = new ImageIcon( currentPath +File.separator +"images"+File.separator+ "favicon.png" ).getImage();
						break;
					case Linux:// Linux用
						image = new ImageIcon( "images" + File.separator + "favicon.png" ).getImage();
						break;
					case Windows:// Windows用
						image = new ImageIcon( "images" + File.separator + "favicon.png" ).getImage();
						break;
					default:
						System.out.println( "Unsupported OS" );
						break;
				}
				TrayIcon icon = new TrayIcon( image, "SKIT", popup );
				icon.setImageAutoSize( true );
				try {
					tray.add( icon );
				} catch ( AWTException _e ) {
					_e.getStackTrace();
				}
			}
		} );
	}
			private enum OS {
		Windows,
		Linux,
		Mac
	}
}