package jp.sagalab.skit;

import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import javax.swing.Timer;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.Matrix;
import jp.sagalab.jftk.transform.TransformMatrix;

/**
 * モデル空間の初期化を行うクラス。
 * @author oyoshi
 */
public class ModelSpaceInitializer implements ActionListener, AnimationData {
	private static final int ANIMATION_FPS = 20;
	private static final int FRAME_NUM = 10;
	private final ReadWriteSKITModel m_model;
	private final Container m_container;
	private MouseListener[] m_mouseListeners;
	private MouseMotionListener[] m_mouseMotionListeners;
	private KeyListener[] m_keyListeners;
	private final Matrix m_initMat;
	private Matrix m_changedMat;
	private final Matrix m_initGridSpaceMat;
	private Matrix m_changedGridSpaceMat;
	private final Timer m_animator;
	private int m_currentFrameIndex = 0;

	public static ModelSpaceInitializer create( ReadWriteSKITModel _model, Container _container ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		if ( _container == null ) {
			throw new NullPointerException( "_container is null." );
		}
		return new ModelSpaceInitializer( _model, _container );
	}

	private ModelSpaceInitializer( ReadWriteSKITModel _model, Container _container ) {
		m_model = _model;
		m_container = _container;
		m_initMat = Matrix.create( m_model.getInitialViewToModel().elements() );
		m_animator = new Timer( (int) Math.round( 1000.0 / ANIMATION_FPS ), this );
		m_initGridSpaceMat = Matrix.create( m_model.getInitialGridSpace().grid().elements() );
	}

	@Override
	public void startAnimation() {
		m_mouseListeners = m_container.getMouseListeners();
		m_mouseMotionListeners = m_container.getMouseMotionListeners();
		m_keyListeners = m_container.getKeyListeners();
		m_changedMat = Matrix.create( m_model.getViewToModel().elements() );
		m_changedGridSpaceMat = Matrix.create( m_model.getGridSpace().grid().elements() );
		removeMouseInputListeners();
		removeKeyListeners();
		m_animator.start();
	}

	@Override
	public void actionPerformed( ActionEvent e ) {
		if ( !updateScene() ) {
			m_animator.stop();
			addMouseInputListeners();
			addKeyListeners();
		}
	}

	@Override
	public boolean updateScene() {
		if ( m_currentFrameIndex < FRAME_NUM ) {
			double weight = m_currentFrameIndex / (double) ( FRAME_NUM - 1 );
			Matrix weightMat0 = createWeightMatix( weight );
			Matrix weightMat1 = createWeightMatix( 1.0 - weight );

			Matrix internalDivisionMatrix = internalDivision( m_changedMat, m_initMat, weightMat0, weightMat1 );
			TransformMatrix mat = TransformMatrix.create(internalDivisionMatrix.elements() );
			m_model.setViewToModel( mat );

			Matrix internalDivisionGridSpaceMat = internalDivision( m_changedGridSpaceMat, m_initGridSpaceMat, weightMat0, weightMat1 );
			GridSpace initialGridSpace = m_model.getInitialGridSpace();
			GridSpace gridSpace = GridSpace.create(TransformMatrix.create(
				internalDivisionGridSpaceMat.elements() ),
				initialGridSpace.basedFuzziness(),
				initialGridSpace.factor() );
			m_model.setGridSpace( gridSpace );

			++m_currentFrameIndex;
			return true;
		} else {
			return false;
		}
	}

	private void removeMouseInputListeners() {
		MouseListener[] mouseListeners = m_container.getMouseListeners();
		MouseMotionListener[] mouseMotionListeners = m_container.getMouseMotionListeners();
		for ( MouseListener mouseListener : mouseListeners ) {
			m_container.removeMouseListener( mouseListener );
		}
		for ( MouseMotionListener mouseMotionListener : mouseMotionListeners ) {
			m_container.removeMouseMotionListener( mouseMotionListener );
		}
	}

	private void removeKeyListeners() {
		KeyListener[] keyListeners = m_container.getKeyListeners();
		for ( KeyListener keyListener : keyListeners ) {
			m_container.removeKeyListener( keyListener );
		}
	}

	private void addMouseInputListeners() {
		for ( MouseListener mouseListener : m_mouseListeners ) {
			m_container.addMouseListener( mouseListener );
		}
		for ( MouseMotionListener mouseMotionListener : m_mouseMotionListeners ) {
			m_container.addMouseMotionListener( mouseMotionListener );
		}
	}

	private void addKeyListeners() {
		for ( KeyListener keyListener : m_keyListeners ) {
			m_container.addKeyListener( keyListener );
		}
	}

	private static Matrix createWeightMatix( double _weight ) {
		Matrix weightMatrix = Matrix.create( new double[][]{
				{ _weight, 0, 0, 0 },
				{ 0, _weight, 0, 0 },
				{ 0, 0, _weight, 0 },
				{ 0, 0, 0, _weight }
			} );
		return weightMatrix;
	}

	private static Matrix internalDivision( Matrix _matA, Matrix _matB, Matrix _ratioMatA, Matrix _ratioMatB ) {
		Matrix internalDivisionMat = _matA.product( _ratioMatB ).plus( _matB.product( _ratioMatA ) );
		return internalDivisionMat;
	}
}
