package jp.sagalab.skit;

import java.awt.event.MouseEvent;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import javax.swing.event.MouseInputAdapter;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.snapper.IFGSSnapper;

/**
 * 絶対図形空間操作のコントローラ。
 *
 * @author oyoshi
 */
public class MouseAbsoluteModelSpaceController extends MouseInputAdapter {
	/** モデル */
	private final ReadWriteSKITModel m_model;
	/** 操作対象となる（モデル空間上の）点列 */
	private final Point[] m_featurePoints;
	/** 固定点のインデックス（m_featurePointsに含まれる点） */
	private int m_pivotIndex;
	/** 選択点のインデックス（m_featurePointsに含まれる点） */
	private int m_selectedIndex;
	/** 絶対図形空間コントローラ */
	private final AbsoluteModelSpaceController m_absoluteModelSpaceController;

	public static MouseAbsoluteModelSpaceController create( ReadWriteSKITModel _model ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		Point[] featurePoints = createFeaturePoints( _model );
		if ( featurePoints.length < 2 ) {
			throw new IllegalStateException( "Length of featurePoints < 2" );
		}
		return new MouseAbsoluteModelSpaceController( _model, featurePoints );
	}

	private MouseAbsoluteModelSpaceController( ReadWriteSKITModel _model, Point[] _featurePoints ) {
		m_model = _model;
		m_featurePoints = _featurePoints;
		m_pivotIndex = -1;
		m_selectedIndex = -1;
		m_absoluteModelSpaceController = AbsoluteModelSpaceController.create( _model );
		_model.setFeaturePoints( m_featurePoints );
	}

	@Override
	public void mousePressed( MouseEvent me ) {
		// モデル空間への変換行列
		TransformMatrix mat = m_model.getViewToModel();
		// Viewport空間への変換行列
		TransformMatrix invMat = mat.inverse();
		// 現在の点
		Point current = Point.createXYZ( me.getX(), me.getY(), 0.0 );
		// 前回の選択点
		int preSelected = m_selectedIndex;
		// 今回の選択点
		int selectedIndex = searchNearestFeaturePoint( current, null );
		// 固定点
		int pivotIndex = m_pivotIndex;
		if ( pivotIndex == -1 ) {
			// 固定点が未決定の場合、選択点から最も遠い点を固定点にする
			Point p = m_featurePoints[selectedIndex].transform( invMat );
			pivotIndex = searchFarthestFeaturePoint( p, m_featurePoints[selectedIndex] );
		} else if ( selectedIndex != preSelected ) {
			// 前回と違う点を選択した場合、前回の選択点を固定点にする
			pivotIndex = preSelected;
		}

		m_absoluteModelSpaceController.move( m_featurePoints[pivotIndex], m_featurePoints[selectedIndex], current.transform( mat ) );

		// 固定点と選択点を決定
		m_pivotIndex = pivotIndex;
		m_selectedIndex = selectedIndex;
		m_model.setRegardPointIndexes( m_pivotIndex, m_selectedIndex );
	}

	@Override
	public void mouseMoved( MouseEvent me ) {
		// マウスが押されていない、かつ動いた時に呼ばれます。
		// マウスが押される前に予め選択点、固定点をセットすることで
		// 「今、マウスを押したらどの点が選択点となるか」をユーザに示しています。

		// モデル空間への変換行列
		TransformMatrix mat = m_model.getViewToModel();
		// Viewport空間への変換行列
		TransformMatrix invMat = mat.inverse();
		// 現在の点
		Point current = Point.createXYZ( me.getX(), me.getY(), 0.0 );
		// 前回の選択点
		int preSelected = m_selectedIndex;
		// 今回の選択点
		int selectedIndex = searchNearestFeaturePoint( current, null );
		// 固定点
		int pivotIndex = m_pivotIndex;
		if ( pivotIndex == -1 ) {
			// 固定点が未決定の場合、選択点から最も遠い点を固定点にする
			Point p = m_featurePoints[selectedIndex].transform( invMat );
			pivotIndex = searchFarthestFeaturePoint( p, m_featurePoints[selectedIndex] );
		} else if ( selectedIndex != preSelected ) {
			// 前回と違う点を選択した場合、前回の選択点を固定点にする
			pivotIndex = preSelected;
		}
		m_model.setRegardPointIndexes( pivotIndex, selectedIndex );

	}

	@Override
	public void mouseDragged( MouseEvent me ) {
		// 移動先の点
		Point current = Point.createXYZ( me.getX(), me.getY(), 0.0 );
		if ( m_pivotIndex == -1 || m_selectedIndex == -1 ) {
			return;
		}
		// モデル空間への変換行列
		TransformMatrix mat = m_model.getViewToModel();
		m_absoluteModelSpaceController.move( m_featurePoints[m_pivotIndex], m_featurePoints[m_selectedIndex], current.transform( mat ) );
	}

	@Override
	public void mouseReleased( MouseEvent me ) {
		if ( m_pivotIndex == -1 || m_selectedIndex == -1 ) {
			return;
		}

		GridSpace gridSpace = m_model.getGridSpace();
		// モデル空間への変換行列
		TransformMatrix mat = m_model.getViewToModel();
		Point current = Point.createXYZ( me.getX(), me.getY(), 0.0 );
		// 移動先の点をスナッピング
		Point point = IFGSSnapper.snapBasedGrid( gridSpace, current.transform( mat ) );

		m_absoluteModelSpaceController.move( m_featurePoints[m_pivotIndex], m_featurePoints[m_selectedIndex], point );
	}

	/**
	 * 指定された点からViewport空間上で最も遠い点を特徴点列の中から探索して返す。
	 *
	 * @param _point     Viewport空間上の点。
	 * @param _exclusion 探索対象から除外する（モデル空間上の）特徴点。
	 * @return Viewport空間上で最も遠い（モデル空間上の）特徴点。
	 */
	private int searchFarthestFeaturePoint( Point _point, Point _exclusion ) {
		double x = _point.x();
		double y = _point.y();
		double z = _point.z();

		// モデル空間からViewport空間への変換行列
		TransformMatrix mat = m_model.getViewToModel().inverse();

		int innermostPoint = -1;
		double maxDistance = Double.NEGATIVE_INFINITY;
		for ( int i = 0; i < m_featurePoints.length; ++i ) {
			if ( _exclusion == null || m_featurePoints[i] != _exclusion ) {
				Point p = m_featurePoints[i].transform( mat );
				double distX = p.x() - x;
				double distY = p.y() - y;
				double distZ = p.z() - z;
				double squaredDistance = distX * distX + distY * distY + distZ * distZ;
				if ( squaredDistance > maxDistance ) {
					innermostPoint = i;
					maxDistance = squaredDistance;
				}
			}
		}

		return innermostPoint;
	}

	/**
	 * 指定された点からViewport空間上で最も近い点を探索して返す。
	 *
	 * @param _point     Viewport空間上の点。
	 * @param _exclusion 探索対象から除外する（モデル空間上の）特徴点。
	 * @return Viewport空間上で最も近い（モデル空間上の）特徴点。
	 */
	private int searchNearestFeaturePoint( Point _point, Point _exclusion ) {
		double x = _point.x();
		double y = _point.y();
		double z = _point.z();

		// モデル空間からViewport空間への変換行列
		TransformMatrix mat = m_model.getViewToModel().inverse();

		int nearestPoint = -1;
		double minDistance = Double.POSITIVE_INFINITY;
		for ( int i = 0; i < m_featurePoints.length; ++i ) {
			if ( _exclusion == null || m_featurePoints[i] != _exclusion ) {
				Point p = m_featurePoints[i].transform( mat );
				double distX = p.x() - x;
				double distY = p.y() - y;
				double distZ = p.z() - z;
				double squaredDistance = distX * distX + distY * distY + distZ * distZ;
				if ( squaredDistance < minDistance ) {
					nearestPoint = i;
					minDistance = squaredDistance;
				}
			}
		}

		return nearestPoint;
	}

	private static Point[] createFeaturePoints( ReadWriteSKITModel _model ) {
		List<Shape> shapes = _model.getShapes();
		Set<Point> featurePoints = new HashSet<Point>( shapes.size() * 2 );
		for ( Shape s : shapes ) {
			featurePoints.addAll( Arrays.asList( s.featurePoints() ) );
		}
		return featurePoints.toArray( new Point[ featurePoints.size() ] );
	}
}
