package jp.sagalab.skit;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.util.List;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jftk.curve.primitive.PrimitiveCurve;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.Point;

/**
 * プリミティブ曲線描画クラス
 *
 * @author kaneko
 */
public class PrimitiveCurveLayer extends Component {
	private static final Color HOT_PRIMITIVE_COLOR = new Color( 214, 247, 59, 255 );
	private static final Stroke HOT_PRIMITIVE_STROKE = new BasicStroke( 2.5f );
	private static final Color ACTIVE_PRIMITIVE_COLOR = new Color( 99, 191, 17, 190 );
	private static final Stroke ACTIVE_PRIMITIVE_STROKE = new BasicStroke( 2.5f );
	private static final Color INACTIVE_PRIMITIVE_COLOR = new Color( 230, 230, 230, 255 );
	private static final Stroke INACTIVE_PRIMITIVE_STROKE = new BasicStroke( 2.5f );
	private static final Color PRIMITIVE_SHADOW_COLOR = new Color( 63, 63, 63, 255 );
	private static final BasicStroke DOTTED_LINE_STROKE = new BasicStroke( (float) 1.0, BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER, 10.0f, new float[]{ 2.0f, 3.0f }, 0.0f );
	private static final Font FONT = new Font( "serif", Font.BOLD, 24 );
	private final ReadOnlySKITModel m_model;

	/**
	 * レイヤーの生成
	 *
	 * @param _model
	 * @param _width
	 * @param _height
	 * @return インスタンス
	 */
	public static PrimitiveCurveLayer create( ReadOnlySKITModel _model, int _width, int _height ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		if ( _width <= 0 ) {
			throw new IllegalArgumentException( "_width <= 0" );
		}
		if ( _height <= 0 ) {
			throw new IllegalArgumentException( "_height <= 0" );
		}
		return new PrimitiveCurveLayer( _model, _width, _height );
	}

	private PrimitiveCurveLayer( ReadOnlySKITModel _model, int _width, int _height ) {
		m_model = _model;
		setBackground( new Color( 0, 0, 0, 0 ) );
		setSize( _width, _height );
	}

	@Override
	public void paint( Graphics _g ) {
		Graphics2D g = (Graphics2D) _g;

		TransformMatrix matModelToView = m_model.getViewToModel().inverse();

		// 非アクティブな図形群の描画
		drawInactiveShapes( g, m_model.getInactiveShapes(), matModelToView );
		// アクティブな図形群の描画
		drawActiveShapes( g, m_model.getActiveShapes(), matModelToView );
		// ホットな図形群の描画
		drawHotShapes( g, m_model.getHotShapes(), matModelToView );
	}

	/**
	 * 非アクティブ図形群を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _shapes 図形群
	 * @param _matModelToView モデル空間からビューポート空間への変換行列
	 */
	private static void drawInactiveShapes( Graphics2D _g, List<Shape> _shapes, TransformMatrix _matModelToView ) {
		for ( Shape s : _shapes ) {
			drawInactivePrimitiveCurves( _g, s.primitives(), _matModelToView );
		}
	}

	/**
	 * アクティブ図形群を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _shapes 図形群
	 * @param _matModelToView モデル空間からビューポート空間への変換行列
	 */
	private static void drawActiveShapes( Graphics2D _g, List<Shape> _shapes, TransformMatrix _matModelToView ) {
		// 図形群描画
		for ( Shape s : _shapes ) {
			// プリミティブ列描画
			drawActivePrimitiveCurves( _g, s.primitives(), _matModelToView );
		}
	}

	/**
	 * ホット図形群を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _shapes 図形群
	 * @param _matModelToView モデル空間からビューポート空間への変換行列
	 */
	private static void drawHotShapes( Graphics2D _g, List<Shape> _shapes, TransformMatrix _matModelToView ) {
		// 図形群描画
		for ( Shape s : _shapes ) {
			// プリミティブ列描画
			drawHotPrimitiveCurves( _g, s.primitives(), _matModelToView );
		}
	}

	/**
	 * 非アクティブなプリミティブ曲線列の描画
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _primitives プリミティブ曲線列
	 * @param _modelToView モデル空間からビューポート空間への変換行列
	 */
	private static void drawInactivePrimitiveCurves( Graphics2D _g, List<PrimitiveCurve> _primitives, TransformMatrix _matModelToView ) {
		_g.setColor( INACTIVE_PRIMITIVE_COLOR );
		_g.setStroke( INACTIVE_PRIMITIVE_STROKE );
		for ( PrimitiveCurve primitive : _primitives ) {
			PrimitiveCurve transed = primitive.transform( _matModelToView );
			// プリミティブ描画
			SKITPanelHelper.drawPrimitive( _g, transed );
		}
	}

	/**
	 * アクティブなプリミティブ曲線列の描画
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _primitives プリミティブ曲線列
	 * @param _modelToView モデル空間からビューポート空間への変換行列
	 */
	private static void drawActivePrimitiveCurves( Graphics2D _g, List<PrimitiveCurve> _primitives, TransformMatrix _matModelToView ) {
		int times = _primitives.size() - 1;
		Point otherEndPoint = null;

		_g.setColor( ACTIVE_PRIMITIVE_COLOR );

		for ( int i = 0; i < times; ++i ) {
			PrimitiveCurve transed = _primitives.get( i ).transform( _matModelToView );
			//曲線列の内、直前のプリミティブの終点がある場合は連結線を描画
			if ( otherEndPoint != null ) {
				Point startPoint = transed.locus( transed.range().start() );
				double strokeWidth = 1.0;
				//ミシン目
				_g.setStroke( DOTTED_LINE_STROKE );
				SKITPanelHelper.drawLine( _g, otherEndPoint, startPoint );
			}

			_g.setStroke( ACTIVE_PRIMITIVE_STROKE );
			// プリミティブ描画
			SKITPanelHelper.drawPrimitive( _g, transed );
			// プリミティブの始点と終点を強調
			SKITPanelHelper.fillCircle( _g, transed.locus( transed.range().start() ), 4.0 );
			SKITPanelHelper.fillCircle( _g, transed.locus( transed.range().end() ), 4.0 );
			otherEndPoint = transed.locus( transed.range().end() );
		}

		// 末尾のプリミティブ取得
		PrimitiveCurve transed = _primitives.get( times ).transform( _matModelToView );
		//曲線列の内、直前のプリミティブの終点がある場合は連結線を描画
		if ( otherEndPoint != null ) {
			Point startPoint = transed.locus( transed.range().start() );
			double strokeWidth = 1.0;
			//ミシン目
			_g.setStroke( DOTTED_LINE_STROKE );
			SKITPanelHelper.drawLine( _g, otherEndPoint, startPoint );
		}

		_g.setStroke( ACTIVE_PRIMITIVE_STROKE );
		// 末尾のプリミティブ描画
		SKITPanelHelper.drawPrimitive( _g, transed );
		// プリミティブの始点を強調
		SKITPanelHelper.fillCircle( _g, transed.locus( transed.range().start() ), 4.0 );
		// プリミティブの終点に矢尻を描画
		SKITPanelHelper.drawArrowHeadOfCurve( _g, transed );
	}

	/**
	 * ホットなプリミティブ曲線列の描画
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _primitives プリミティブ曲線列
	 * @param _modelToView モデル空間からビューポート空間への変換行列
	 */
	private static void drawHotPrimitiveCurves( Graphics2D _g, List<PrimitiveCurve> _primitives, TransformMatrix _matModelToView ) {
		int times = _primitives.size() - 1;
		Point otherEndPoint = null;

		for ( int i = 0; i < times; ++i ) {
			PrimitiveCurve transed = _primitives.get( i ).transform( _matModelToView );
			_g.setColor( HOT_PRIMITIVE_COLOR );
			//曲線列の内、直前のプリミティブの終点がある場合は連結線を描画
			if ( otherEndPoint != null ) {
				Point startPoint = transed.locus( transed.range().start() );
				double strokeWidth = 1.0;
				//ミシン目
				_g.setStroke( DOTTED_LINE_STROKE );
				SKITPanelHelper.drawLine( _g, otherEndPoint, startPoint );
			}

			_g.setStroke( HOT_PRIMITIVE_STROKE );
			// プリミティブ描画
			SKITPanelHelper.drawPrimitive( _g, transed );
			// プリミティブの始点と終点を強調
			SKITPanelHelper.fillCircle( _g, transed.locus( transed.range().start() ), 4.0 );
			SKITPanelHelper.fillCircle( _g, transed.locus( transed.range().end() ), 4.0 );
			// プリミティブラベル描画
			SKITPanelHelper.drawPrimitiveType( _g, transed, FONT, HOT_PRIMITIVE_COLOR, PRIMITIVE_SHADOW_COLOR );
			otherEndPoint = transed.locus( transed.range().end() );
		}

		// 末尾のプリミティブ取得
		PrimitiveCurve transed = _primitives.get( times ).transform( _matModelToView );
		//曲線列の内、直前のプリミティブの終点がある場合は連結線を描画
		if ( otherEndPoint != null ) {
			Point startPoint = transed.locus( transed.range().start() );
			double strokeWidth = 1.0;
			//ミシン目
			_g.setStroke( DOTTED_LINE_STROKE );
			SKITPanelHelper.drawLine( _g, otherEndPoint, startPoint );
		}

		_g.setColor( HOT_PRIMITIVE_COLOR );
		_g.setStroke( HOT_PRIMITIVE_STROKE );
		// 末尾のプリミティブ描画
		SKITPanelHelper.drawPrimitive( _g, transed );
		// プリミティブの始点を強調
		SKITPanelHelper.fillCircle( _g, transed.locus( transed.range().start() ), 4.0 );
		// プリミティブの終点に矢尻を描画
		SKITPanelHelper.drawArrowHeadOfCurve( _g, transed );
		// プリミティブラベル描画
		SKITPanelHelper.drawPrimitiveType( _g, transed, FONT, HOT_PRIMITIVE_COLOR, PRIMITIVE_SHADOW_COLOR );
	}
}
