package jp.sagalab.skit;

import jp.sagalab.jftk.snapper.IFGSSnapper;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.jftk.Vector;

/**
 * 相対図形空間操作のコントローラ。
 * @author ishiguro
 */
public class RelativeModelSpaceController {
	/** モデル空間 */
	private final ReadWriteSKITModel m_model;
	/** グリッドの原点 */
	private final Point m_originOfGridSpace;
	/** 拡縮倍率 XXX 2の10乗に決め打ち */
	private final static double m_scale = 1024;

	/**
	 * 相対図形空間操作のコントローラの生成。
	 * @param _model モデル空間。
	 * @param _mi モデル空間情報。
	 * @param _vi ビューポート情報。
	 * @return コントローラ。
	 */
	public static RelativeModelSpaceController create( ReadWriteSKITModel _model, ModelSpaceInfo _mi, ViewportInfo _vi ) {
		if ( _model == null ) {
			throw new NullPointerException( "_model is null." );
		}
		if ( _mi == null ) {
			throw new NullPointerException( "_mi is null." );
		}
		if ( _vi == null ) {
			throw new NullPointerException( "_vi is null." );
		}
		// 初期のビューポート空間からモデル空間への変換行列を取得
		TransformMatrix initViewToModel = _model.getInitialViewToModel();
		// ビューポートの中心を求める
		Point centerOfViewport = Point.createXYZ( _vi.width() / 2.0, _vi.height() / 2.0, 0.0 );
		Point originOfGridSpace = IFGSSnapper.snapBasedGrid(_model.getInitialGridSpace(), centerOfViewport.transform( _mi.mappingMatrix()) );
		originOfGridSpace = originOfGridSpace.transform( initViewToModel.inverse() );

		return new RelativeModelSpaceController( _model, originOfGridSpace );
	}

	/**
	 * コンストラクタ。
	 * @param _model モデル空間。
	 * @param _origin グリッドの原点。
	 */
	private RelativeModelSpaceController( ReadWriteSKITModel _model, Point _origin ) {
		m_model = _model;
		m_originOfGridSpace = _origin;
	}

	/**
	 * モデルを移動させる。(モデル空間単位)
	 * @param _x x軸方向への移動量(モデル空間単位)。
	 * @param _y y軸方向への移動量(モデル空間単位)。
	 * @param _z z軸方向への移動量(モデル空間単位)。
	 */
	public void move( double _x, double _y, double _z ) {
		double modelScale = m_model.getViewToModel().scalalize();
		double gridScale = m_model.getGridSpace().grid().scalalize();
		TransformMatrix transModelSpace = TransformMatrix.translation( _x/modelScale, _y/modelScale, _z/modelScale);
		TransformMatrix transGridSpace = TransformMatrix.translation( _x/gridScale, _y/gridScale, _z/gridScale);
		TransformMatrix tmp = m_model.getViewToModel();

		if ( updateTransform( transModelSpace ) ) {
			if ( !updateGridSpace( transGridSpace ) ) {
				//グリッド空間の更新に失敗した場合モデル空間を元の状態に戻す
				m_model.setViewToModel( tmp );
			}
		}
	}

	/**
	 * モデルを回転させる。
	 * @param _rate 回転量(ラジアン)。
	 */
	public void rotate( double _rate ) {
		Vector AXIS_Z = Vector.createXYZ( 0, 0, 1 );
		TransformMatrix rotateModelSpace = TransformMatrix.identity().rotate( _rate, AXIS_Z, m_originOfGridSpace );
		TransformMatrix rotateGridSpace = TransformMatrix.rotation(_rate, AXIS_Z );
		TransformMatrix tmp = m_model.getViewToModel();

		if ( updateTransform( rotateModelSpace ) ) {
			if ( !updateGridSpace( rotateGridSpace ) ) {
				//グリッド空間の更新に失敗した場合モデル空間を元の状態に戻す
				m_model.setViewToModel( tmp );
			}
		}
	}

	/**
	 * モデルを拡大・縮小する。 拡大・縮小率は正の実数で、1より大きければ縮小、1より小さければ拡大を表す。
	 * @param _rate 拡大・縮小率。
	 */
	public void scale( double _rate ) {
		// 拡大・縮小率が0以下の場合は1.0倍とする
		double rate = _rate <= 0 ? 1.0 : _rate;
		TransformMatrix scaleModelSpace = TransformMatrix.identity().scale( rate, m_originOfGridSpace );
		TransformMatrix scaleGridSpace = TransformMatrix.scaling(rate);
		TransformMatrix tmp = m_model.getViewToModel();

		if ( updateTransform( scaleModelSpace ) ) {
			if ( !updateGridSpace( scaleGridSpace ) ) {
				//グリッド空間の更新に失敗した場合モデル空間を元の状態に戻す
				m_model.setViewToModel( tmp );
			}
		}
	}

	/**
	 * モデル空間への変換行列の更新。
	 * @param _mat モデル空間上での平行移動、スケーリング、回転を表す変換行列。
	 */
	private boolean updateTransform( TransformMatrix _mat ) {
		TransformMatrix mat = _mat.product( m_model.getViewToModel() );
		double min = 1/m_scale;
		double max = m_scale;
		if ( mat.scalalize() >= min && mat.scalalize() <= max ) {
			m_model.setViewToModel( mat );
			return true;
		}
		return false;
	}

	/**
	 * グリッド空間の更新。
	 * @param _mat モデル空間上での平行移動、スケーリング、回転を表す変換行列。
	 */
	private boolean updateGridSpace( TransformMatrix _mat ) {
		double scalalize = _mat.scalalize();
		GridSpace grid = m_model.getGridSpace();
		TransformMatrix mat = _mat.product( grid.grid() );

		GridSpace gridSpace = GridSpace.create(mat,
			scalalize * grid.basedFuzziness(), grid.factor() );

		double min = 1/m_scale*1e-3;
		double max = m_scale*1e3;
		if ( mat.scalalize() >= min && mat.scalalize() <= max ) {
			m_model.setGridSpace( gridSpace );
			return true;
		}
		return false;
	}
}
