package jp.sagalab.skit;

import java.util.concurrent.CopyOnWriteArrayList;
import jp.sagalab.jflib.GestureEventDispatcher;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;

/**
 * SKITModelのイベントを配信するクラス。
 */
public class SKITEventDispatcher extends GestureEventDispatcher {
	/** このディスパッチャーが配信するイベントを受信するイベントリスナーのリスト。 */
	private final CopyOnWriteArrayList<SKITEventListener> m_listeners = new CopyOnWriteArrayList<SKITEventListener>();

	/**
	 * 登録されているイベントリスナーの一覧を返す。
	 * @return 登録されているイベントリスナーの一覧。
	 */
	public SKITEventListener[] getShapeListeners() {
		return m_listeners.toArray( new SKITEventListener[ m_listeners.size() ] );
	}

	/**
	 * 新たにイベントリスナーを追加する。
	 * @param _listener この配信者からのイベントを受け取るリスナー。
	 */
	public void addSKITEventListener( SKITEventListener _listener ) {
		addGestureListener( _listener);
		m_listeners.add( _listener );
	}

	/**
	 * 指定したリスナーの登録を解除する。
	 * @param _listener 登録を解除するリスナー。
	 */
	public void removeSKITEventListener( SKITEventListener _listener ) {
		removeGestureListener( _listener);
		m_listeners.remove( _listener );
	}

	/**
	 * 指定したイベントをリスナーに配信する。
	 * @param _added 追加アクティブ図形群。
	 * @param _removed 削除アクティブ図形群。
	 */
	void fireActiveShapesEvent( Shape[] _added, Shape[] _removed ) {
		for ( SKITEventListener l : m_listeners ) {
			l.activeShapesChanged( _added.clone(), _removed.clone() );
		}
	}

	/**
	 * 指定したイベントをリスナーに配信する。
	 * @param _added 追加非アクティブ図形群。
	 * @param _removed 削除非アクティブ図形群。
	 */
	void fireInactiveShapesEvent( Shape[] _added, Shape[] _removed ) {
		for ( SKITEventListener l : m_listeners ) {
			l.inactiveShapesChanged( _added.clone(), _removed.clone() );
		}
	}

	/**
	 * 指定したグリッド空間をリスナーに配信する。
	 * @param _grid 配信するグリッド空間。
	 */
	void fireGridEvent( GridSpace _grid ) {
		for ( SKITEventListener l : m_listeners ) {
			l.gridSpaceChanged( _grid );
		}
	}

	/**
	 * 指定した変換行列をリスナーに配信する。
	 * @param _mat 配信する変換行列。
	 */
	void fireTransformEvent( TransformMatrix _mat ) {
		for ( SKITEventListener l : m_listeners ) {
			l.transformChanged( _mat );
		}
	}
	
	/**
	 * 指定したイベントをリスナーに配信する。
	 * @param _featurePoints 特徴点列。
	 */
	void fireFeaturePointsEvend( Point[] _featurePoints){
		for ( SKITEventListener l : m_listeners ) {
			l.featurePointsChanged( _featurePoints );
		}
	}
	
	/**
	 * 指定したイベントをリスナーに配信する。
	 * @param _pivotIndex 固定点のインデックス。
	 * @param _selectedIndex 選択点のインデックス。
	 */
	void fireRegardPointIndexEvent( int _pivotIndex, int _selectedIndex ){
		for ( SKITEventListener l : m_listeners ) {
			l.regardPointIndexChanged( _pivotIndex, _selectedIndex );
		}
	}
}
