package jp.sagalab.skit;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import jp.sagalab.jflib.Command;
import jp.sagalab.jflib.FSCITask;
import jp.sagalab.jflib.GestureEvent;
import jp.sagalab.jflib.GestureRecognizer;
import jp.sagalab.jflib.History;
import jp.sagalab.jflib.PickingJudge;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jflib.parameter.FSCIParameter;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.curve.SplineCurve;
import jp.sagalab.jftk.fragmentation.Fragment;
import jp.sagalab.jftk.fragmentation.IdentificationFragment;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.transform.TransformMatrix;

/**
 * SKITの根幹をなす、MVCのMに相当するクラス。
 *
 * このクラスはシングルスレッドで操作することを想定して設計されており、非スレッドセーフ。
 */
public class SKITModel extends SKITEventDispatcher implements ReadWriteSKITModel, InputOnlySKITModel {
	/** 図形データベース */
	private final ShapeDatabase m_database;
	/** コマンド履歴 */
	private final History m_history;
	/** FSCI */
	private final FSCITask m_fsci;
	/** ジェスチャ認識 */
	private final GestureRecognizer m_recog;
	/** 初期のビューポート空間からモデル空間への変換行列 */
	private final TransformMatrix m_initialMatViewToModel;
	/** 初期のグリッド空間 */
	private final GridSpace m_initialGridSpace;
	/** ビューポート空間からモデル空間への変換行列 */
	private TransformMatrix m_matViewToModel;
	/** グリッド空間 */
	private GridSpace m_gridSpace;
	/** 特徴点列 */
	private Point[] m_featurePoints;
	/** 絶対図形空間操作における固定点 */
	private int m_pivotIndex = -1;
	/** 絶対図形空間操作における選択点 */
	private int m_selectedIndex = -1;

	/**
	 * グリッド空間を指定してSKITModelを生成する。
	 * @param _viewToModel
	 * @param _gridSpace
	 * @param _ppi         1インチあたりのピクセル数(pixel per inch)。
	 * @param _lei         学習の入力データファイルパス。
	 * @param _fnn         学習の推論ルールファイルパス。
	 * @return 指定したグリッド空間をもつSKTIModel。
	 */
	public static SKITModel create( TransformMatrix _viewToModel, GridSpace _gridSpace,
		double _ppi, String _lei, String _fnn ) {
		FSCITask fsci;
		FSCIParameter parameter = new FSCIParameter.Builder().build( 1.0 / (_ppi / 2.54));
		fsci = FSCITask.create( parameter );

		ShapeDatabase database = ShapeDatabase.create();
		History history = History.create( 20 );
		GestureRecognizer recog = GestureRecognizer.create("/Library/SKIT/etc/gesture.model");
		return new SKITModel( fsci, database, history, recog, _viewToModel, _gridSpace );
	}

	/**
	 * 各パラメータを指定するコンストラクタ。
	 * @param _fsci        FSCI。
	 * @param _database    図形データベース。
	 * @param _history     コマンド履歴。
	 * @param _viewToModel
	 * @param _gridSpace
	 * @param _recog       ジェスチャ認識。
	 */
	protected SKITModel( FSCITask _fsci, ShapeDatabase _database, History _history, GestureRecognizer _recog, TransformMatrix _viewToModel, GridSpace _gridSpace ) {
		m_fsci = _fsci;
		m_database = _database;
		m_history = _history;
		m_recog = _recog;
		m_initialMatViewToModel = m_matViewToModel = _viewToModel;
		m_initialGridSpace = m_gridSpace = _gridSpace;
	}

	@Override
	public void putPoints( Point[] _points ) {
		synchronized ( this ) {
			// 2次ベジェ曲線で評価可能かチェック
			if ( _points.length <= 2 ) {
				return;
			}
			try {
				SplineCurve fsc = m_fsci.createFsc( _points );
				if ( fsc != null ) {
					// 時間伸縮が終わったのでモデル空間上の位置に変換
					SplineCurve transformedFSC = fsc.transform( m_matViewToModel );
					// 書描弁別
					if ( m_fsci.isDrawing( transformedFSC ) ) {
						List<IdentificationFragment> targetList = m_fsci.createBlendTargetList( m_database.hotAndActiveShapes(), transformedFSC, 0.5 );
						final List<Shape> added;
						final List<Shape> removed;
						// 融合対象がある場合
						if ( targetList.size() > 0 ) {
							// アクティブ図形に入力FSCを重ね書きした結果を取得
							final FSCITask.UpdatingDataSet result = m_fsci.overSketching( m_database.hotAndActiveShapes(), targetList, transformedFSC, m_gridSpace );
							added = result.addedShapes();
							removed = result.removedShapes();
						} else {
							List<Fragment> fragments = Arrays.asList( m_fsci.fragmentation( transformedFSC ) );
							Shape recognize = m_fsci.identify(fragments, m_gridSpace );
							List<Shape> recognizes = new ArrayList<Shape>();
							recognizes.add( recognize );
							added = recognizes;
							removed = new ArrayList<Shape>();
						}
						// 重ね書きの結果をコマンドとして実行
						m_history.execute( new Command() {
							@Override
							public void execute() {
								m_database.addAndRemoveActiveShapes( added, removed );
								fireActiveShapesEvent( added.toArray( new Shape[ added.size() ] ), removed.toArray( new Shape[ removed.size() ] ) );
							}

							@Override
							public void unexecute() {
								m_database.addAndRemoveActiveShapes( removed, added );
								fireActiveShapesEvent( removed.toArray( new Shape[ removed.size() ] ), added.toArray( new Shape[ added.size() ] ) );
							}
						} );
					} else {
						// ジェスチャ認識
						GestureEvent gesture = m_recog.recognize( fsc, _points );
						// 各ジェスチャに対応する処理
						switch ( gesture.type() ) {
							case L_SCRATCH_R:
							case R_SCRATCH_R:
							case L_SCRATCH_DR:
							case R_SCRATCH_DR:
							case L_SCRATCH_D:
							case R_SCRATCH_D:
							case L_SCRATCH_DL:
							case R_SCRATCH_DL:
							case L_SCRATCH_L:
							case R_SCRATCH_L:
							case L_SCRATCH_UL:
							case R_SCRATCH_UL:
							case L_SCRATCH_U:
							case R_SCRATCH_U:
							case L_SCRATCH_UR:
							case R_SCRATCH_UR:
								// 削除
								System.out.println( "scratch" );
								handleScratchGesture( transformedFSC );
								break;
							case SEPARATE:
								// 分離
								System.out.println( "tick" );
								handleTickGesture( transformedFSC );
								break;
							case UNDO:
								// アンドゥ
								System.out.println( "undo" );
								undo();
								break;
							case REDO:
								// リドゥ
								System.out.println( "redo" );
								redo();
								break;
							case F_REVERCE_L:
							case R_REVERCE_L:
							case F_REVERCE_R:
							case R_REVERCE_R:
							case F_REVERCE_U:
							case R_REVERCE_U:
							case F_REVERCE_D:
							case R_REVERCE_D:
							case F_REVERCE_UL:
							case R_REVERCE_UL:
							case F_REVERCE_DR:
							case R_REVERCE_DR:
							case F_REVERCE_DL:
							case R_REVERCE_DL:
							case F_REVERCE_UR:
							case R_REVERCE_UR:
								System.out.println( "slash" );
								handleSlashGesture( transformedFSC );
								break;
							default:
								System.out.println( "unknown" );
								break;
						}
						// ジェスチャイベントを配信
						fireGestureEvent( gesture );
					}
				}
			} catch ( Exception e ) {
				//入力点列の記録
				PointIOUtility.writePoints( _points );
				Logger.getLogger( FSCITask.class.getName() ).log( Level.SEVERE, null, e );

			}
		}
	}

	private void handleTickGesture( SplineCurve _fsc ) {
		FSCITask.UpdatingDataSet result = FSCITask.separate( m_database.hotAndActiveShapes(), _fsc, 0.5 );
		final List<Shape> added = result.addedShapes();
		final List<Shape> removed = result.removedShapes();
		if ( !removed.isEmpty() ) {
			m_history.execute( new Command() {
				@Override
				public void execute() {
					m_database.addAndRemoveActiveShapes( added, removed );
					fireActiveShapesEvent( added.toArray( new Shape[ added.size() ] ), removed.toArray( new Shape[ removed.size() ] ) );
				}

				@Override
				public void unexecute() {
					m_database.addAndRemoveActiveShapes( removed, added );
					fireActiveShapesEvent( removed.toArray( new Shape[ removed.size() ] ), added.toArray( new Shape[ added.size() ] ) );
				}
			} );
		}
	}

	private void handleSlashGesture( SplineCurve _fsc ) {
		FSCITask.UpdatingDataSet result = FSCITask.invert( m_database.hotAndActiveShapes(), _fsc, 0.5 );
		final List<Shape> added = result.addedShapes();
		final List<Shape> removed = result.removedShapes();
		if ( !removed.isEmpty() ) {
			m_history.execute( new Command() {
				@Override
				public void execute() {
					m_database.addAndRemoveActiveShapes( added, removed );
					fireActiveShapesEvent( added.toArray( new Shape[ added.size() ] ), removed.toArray( new Shape[ removed.size() ] ) );
				}

				@Override
				public void unexecute() {
					m_database.addAndRemoveActiveShapes( removed, added );
					fireActiveShapesEvent( removed.toArray( new Shape[ removed.size() ] ), added.toArray( new Shape[ added.size() ] ) );
				}
			} );
		}
	}

	private void handleScratchGesture( SplineCurve fsc ) {
		FSCITask.UpdatingDataSet result = FSCITask.remove( m_database.hotAndActiveShapes(), fsc, 0.5 );
		final List<Shape> added = result.addedShapes();
		final List<Shape> removed = result.removedShapes();
		if ( !removed.isEmpty() ) {
			m_history.execute( new Command() {
				@Override
				public void execute() {
					m_database.addAndRemoveActiveShapes( added, removed );
					fireActiveShapesEvent( added.toArray( new Shape[ added.size() ] ), removed.toArray( new Shape[ removed.size() ] ) );
				}

				@Override
				public void unexecute() {
					m_database.addAndRemoveActiveShapes( removed, added );
					fireActiveShapesEvent( removed.toArray( new Shape[ removed.size() ] ), added.toArray( new Shape[ added.size() ] ) );
				}
			} );
		}
	}

	@Override
	public List<Shape> getShapes() {
		return m_database.allShapes();
	}

	@Override
	public List<Shape> getHotShapes() {
		return m_database.hotShapes();
	}

	@Override
	public List<Shape> getActiveShapes() {
		return m_database.activeShapes();
	}

	@Override
	public List<Shape> getInactiveShapes() {
		return m_database.inactiveShapes();
	}

	@Override
	public Point[] getFeaturePoints() {
		return m_featurePoints.clone();
	}

	@Override
	public int getPivotPointIndex() {
		return m_pivotIndex;
	}

	@Override
	public int getSelectedPointIndex() {
		return m_selectedIndex;
	}

	@Override
	public TransformMatrix getInitialViewToModel() {
		return m_initialMatViewToModel;
	}

	@Override
	public GridSpace getInitialGridSpace() {
		return m_initialGridSpace;
	}

	@Override
	public TransformMatrix getViewToModel() {
		return m_matViewToModel;
	}

	@Override
	public GridSpace getGridSpace() {
		return m_gridSpace;
	}

	@Override
	public void setViewToModel( TransformMatrix _mat ) {
		m_matViewToModel = _mat;
		fireTransformEvent( _mat );
	}

	@Override
	public void setGridSpace( GridSpace _gridSpace ) {
		m_gridSpace = _gridSpace;
		fireGridEvent( _gridSpace );
	}

	@Override
	public void setFeaturePoints( Point[] _points ) {
		m_featurePoints = _points.clone();
		fireFeaturePointsEvend( m_featurePoints );
	}

	@Override
	public void setRegardPointIndexes( int _pivotIndex, int _selectedIndex ) {
		m_pivotIndex = _pivotIndex;
		m_selectedIndex = _selectedIndex;
		fireRegardPointIndexEvent( _pivotIndex, _selectedIndex );
	}

	@Override
	public void resetFeaturePoints() {
		m_featurePoints = null;
		m_pivotIndex = -1;
		m_selectedIndex = -1;
		fireFeaturePointsEvend( m_featurePoints );
		fireRegardPointIndexEvent( m_pivotIndex, m_selectedIndex );
	}

	/**
	 * ビューポート空間からモデル空間への変換行列が変更されているか。
	 * @return 変更されている場合はtrue。
	 */
	public boolean isChangedViewToModelMatrix() {
		return !m_initialMatViewToModel.equals( m_matViewToModel );
	}

	/**
	 * アンドゥを実行する。
	 */
	public void undo() {
		m_history.undo();
	}

	/**
	 * リドゥを実行する。
	 */
	public void redo() {
		m_history.redo();
	}

	@Override
	public void cancelLearning() {
		m_fsci.cancelLearning();
	}

	@Override
	public void picking( Point _point, double _lpd, double _threshold ) {
		synchronized ( this ) {
			// モデル空間中での入力点を算出
			Point point = _point.transform( m_matViewToModel );
			// モデル空間中でのピッキング有効範囲を算出
			double deviceUnit = _threshold / _lpd;
			double threshold = deviceUnit * m_matViewToModel.scalalize();
			Shape pickedShape = PickingJudge.searchPickedShape( point, m_database.inactiveShapes(), threshold );
			if ( pickedShape != null ) {
				// 近くにあった図形をアクティブ化
				final List<Shape> toActive = new ArrayList<Shape>( 1 );
				toActive.add( pickedShape );
				m_history.execute( new Command() {
					@Override
					public void execute() {
						m_database.addAndRemoveActiveShapes( toActive, Collections.<Shape>emptyList() );
						fireActiveShapesEvent( toActive.toArray( new Shape[ toActive.size() ] ), new Shape[ 0 ] );
						m_database.addAndRemoveInactiveShapes( Collections.<Shape>emptyList(), toActive );
						fireActiveShapesEvent( new Shape[ 0 ], toActive.toArray( new Shape[ toActive.size() ] ) );
					}

					@Override
					public void unexecute() {
						m_database.addAndRemoveInactiveShapes( toActive, Collections.<Shape>emptyList() );
						fireInactiveShapesEvent( toActive.toArray( new Shape[ toActive.size() ] ), new Shape[ 0 ] );
						m_database.addAndRemoveActiveShapes( Collections.<Shape>emptyList(), toActive );
						fireInactiveShapesEvent( new Shape[ 0 ], toActive.toArray( new Shape[ toActive.size() ] ) );
					}
				} );
				return;
			}

			// 近くに図形がなく，かつアクティブ図形がない->全アクティブ
			if ( m_database.hotAndActiveShapes().isEmpty() ) {
				System.out.println( "active shapes is empty. " );
				final List<Shape> toActive = new ArrayList<Shape>( m_database.inactiveShapes() );
				// すべてアクティブ化
				m_history.execute( new Command() {
					@Override
					public void execute() {
						m_database.addAndRemoveActiveShapes( toActive, Collections.<Shape>emptyList() );
						fireActiveShapesEvent( toActive.toArray( new Shape[ toActive.size() ] ), new Shape[ 0 ] );
						m_database.addAndRemoveInactiveShapes( Collections.<Shape>emptyList(), toActive );
						fireInactiveShapesEvent( new Shape[ 0 ], toActive.toArray( new Shape[ toActive.size() ] ) );
					}

					@Override
					public void unexecute() {
						m_database.addAndRemoveInactiveShapes( toActive, Collections.<Shape>emptyList() );
						fireInactiveShapesEvent( toActive.toArray( new Shape[ toActive.size() ] ), new Shape[ 0 ] );
						m_database.addAndRemoveActiveShapes( Collections.<Shape>emptyList(), toActive );
						fireActiveShapesEvent( new Shape[ 0 ], toActive.toArray( new Shape[ toActive.size() ] ) );
					}
				} );
			} else {
				System.out.println( "active shapes is not empty. " );
				final List<Shape> toInactive = new ArrayList<Shape>( m_database.hotAndActiveShapes() );
				// 非アクティブにした図形群のスキルデータを使用して学習
				final List<Shape> doneShapes = m_fsci.learning( toInactive );
				// すべて非アクティブ化
				m_history.execute( new Command() {
					@Override
					public void execute() {
						m_database.addAndRemoveInactiveShapes( doneShapes, Collections.<Shape>emptyList() );
						fireInactiveShapesEvent( doneShapes.toArray( new Shape[ doneShapes.size() ] ), new Shape[ 0 ] );
						m_database.addAndRemoveActiveShapes( Collections.<Shape>emptyList(), toInactive );
						fireActiveShapesEvent( new Shape[ 0 ], toInactive.toArray( new Shape[ toInactive.size() ] ) );
					}

					@Override
					public void unexecute() {
						m_database.addAndRemoveActiveShapes( toInactive, Collections.<Shape>emptyList() );
						fireActiveShapesEvent( toInactive.toArray( new Shape[ toInactive.size() ] ), new Shape[ 0 ] );
						m_database.addAndRemoveInactiveShapes( Collections.<Shape>emptyList(), doneShapes );
						fireInactiveShapesEvent( new Shape[ 0 ], doneShapes.toArray( new Shape[ doneShapes.size() ] ) );
					}
				} );
			}
		}
	}
}
