package jp.sagalab.skit;

import java.awt.AWTException;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import jp.sagalab.jflib.GestureEvent;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jftk.snapper.IFGSSnapper;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.skit.mulchitouchdevice.MultiTouchDeviceClient;
import jp.sagalab.skit.mulchitouchdevice.MultiTouchModelSpaceController;

/**
 * SKITのビュー。
 */
public class SKITPanel extends JPanel {

	/**
	 * パネルを生成する。
	 *
	 * @param _ppi 1インチあたりのピクセル数(pixel per inch)。
	 * @param _server XMLサーバ。
	 * @param _mi モデル空間情報。
	 * @param _vi ビューポート情報。
	 * @return パネル。
	 * @throws AWTException
	 */
	public static SKITPanel create( final double _ppi, final XMLServer _server, ModelSpaceInfo _mi, ViewportInfo _vi ) throws AWTException {

		// ビューポート空間からモデル空間への変換行列
		TransformMatrix viewToModel = _mi.mappingMatrix();
		// グリッド生成
		GridSpace gridSpace = createGridSpace( _mi, _vi );
		// TODO Mainで生成
		// モデル生成
		SKITModel model = SKITModel.create( viewToModel, gridSpace, _ppi, "anchor.lei", "anchor.fnn" );

		// マウスリスナー生成
		InputPointsController mouseController = InputPointsController.create( model, _ppi );

		int width = _vi.width();
		int height = _vi.height();
		// 背景画像生成
		BufferedImage background = ImageCreater.createBackgroundImage( _vi.X(), _vi.Y(), width, height );
		// パネル生成
		SKITPanel panel = new SKITPanel();

		// 各レイヤーを生成
		panel.setPreferredSize( new Dimension( width, height ) );
		panel.setLayout( null );
		GridLayer gridLayer = GridLayer.create( model, width, height );
		FragmentLayer fragmentLayer = FragmentLayer.create( model, width, height );
		PrimitiveCurveLayer primitiveCurveLayer = PrimitiveCurveLayer.create( model, width, height );
		GuideLineLayer guideLineLayer = GuideLineLayer.create( model, width, height );
		SnappingDataLayer snappingDataLayer = SnappingDataLayer.create( model, width, height );
		GestureLayer gestureLayer = GestureLayer.create( width, height );
		ScratchLayer scratchLayer = ScratchLayer.create( width, height );
		BackGroundLayer backGroundLayer = BackGroundLayer.create( background, model );
		// 各レイヤーをパネルに登録
		panel.add( scratchLayer );
		panel.add( gestureLayer );
		panel.add( snappingDataLayer );
		panel.add( primitiveCurveLayer );
		panel.add( guideLineLayer );
		panel.add( fragmentLayer );
		panel.add( gridLayer );
		panel.add( backGroundLayer );

		// マウスリスナー登録
		panel.addMouseListener( mouseController );
		panel.addMouseMotionListener( mouseController );
		panel.addMouseListener( scratchLayer );
		panel.addMouseMotionListener( scratchLayer );

		// キーリスナー登録
		RelativeModelSpaceController relativeModelSpaceController = RelativeModelSpaceController.create( model, _mi, _vi );
		panel.addKeyListener( KeyController.create( model, _server, panel, relativeModelSpaceController, _ppi ) );

		MultiTouchDeviceClient mtdc = new MultiTouchDeviceClient(_vi );
		MultiTouchModelSpaceController mtmsc = new MultiTouchModelSpaceController(model, panel, relativeModelSpaceController );
		mtdc.addDeviceListener( mtmsc );

		// SKITイベントリスナー登録
		model.addSKITEventListener( panel.createSKITEventListener( gestureLayer ) );

		return panel;
	}

	private SKITPanel() {
	}

	@Override
	public void paintComponent( Graphics _g ) {
		Graphics2D g = (Graphics2D) _g;
		g.setBackground( new Color(0.25f,0.25f,0.25f,1));
		g.clearRect( getX(), getY(), getWidth(), getHeight() );
		g.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
	}

	/**
	 * グリッドを生成し、取得する。
	 *
	 * @param _mi モデル空間情報。
	 * @param _vi ビューポート情報。
	 * @return グリッド空間。
	 */
	private static GridSpace createGridSpace( ModelSpaceInfo _mi, ViewportInfo _vi ) {
		// グリッドを生成
		TransformMatrix basedGrid = _mi.gridSpaceMatrix();
		double fuzziness = basedGrid.scalalize() / 4.0;
		GridSpace grid = GridSpace.create(basedGrid, fuzziness, _mi.derivatinalScaleFactor() );
		// ビューポートの中心を求める
		Point center = Point.createXYZ( _vi.width() / 2.0, _vi.height() / 2.0, 0.0 );
		// ビューポートの中心付近のグリッド交点を求める
		center = IFGSSnapper.snapBasedGrid( grid, center.transform( _mi.mappingMatrix() ) );
		// 求めたグリッド交点を原点としてグリッドを再生成
		double[][] elements = basedGrid.elements();
		elements[0][3] = center.x();
		elements[1][3] = center.y();
		elements[2][3] = center.z();
		basedGrid = TransformMatrix.create(elements );
		grid = GridSpace.create(basedGrid, grid.basedFuzziness(), grid.factor() );

		return grid;
	}

	/**
	 * SKITイベントリスナーを生成する。
	 *
	 * @param _gestureLayer
	 * @return
	 */
	private SKITEventListener createSKITEventListener( final GestureLayer _gestureLayer ) {
		return new SKITEventListener() {
			Runnable updater = new Runnable() {
				@Override
				public void run() {
					repaint();
				}
			};

			@Override
			public void activeShapesChanged( Shape[] _added, Shape[] _removed ) {
				SwingUtilities.invokeLater( updater );
			}

			@Override
			public void inactiveShapesChanged( Shape[] _added, Shape[] _removed ) {
				SwingUtilities.invokeLater( updater );
			}

			@Override
			public void gridSpaceChanged( GridSpace _gridSpace ) {
				SwingUtilities.invokeLater( updater );
			}

			@Override
			public void transformChanged( TransformMatrix _mat ) {
				SwingUtilities.invokeLater( updater );
			}

			@Override
			public void receivedGesture( GestureEvent _gesture ) {
				AnimationData data = _gestureLayer.addGesture( _gesture );
				data.startAnimation();
			}

			@Override
			public void featurePointsChanged( Point[] _featurePoints ) {
				SwingUtilities.invokeLater( updater );
			}

			@Override
			public void regardPointIndexChanged( int _pivotIndex, int _selectedIndex ) {
				SwingUtilities.invokeLater( updater );
			}
		};
	}
}
