package jp.sagalab.skit;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.Vector;
import jp.sagalab.jftk.curve.BezierCurve;
import jp.sagalab.jftk.curve.ParametricCurve;
import jp.sagalab.jftk.curve.ParametricEvaluable.EvaluationType;
import jp.sagalab.jftk.curve.Range;
import jp.sagalab.jftk.curve.primitive.Circle;
import jp.sagalab.jftk.curve.primitive.CircularArc;
import jp.sagalab.jftk.curve.primitive.ClosedFreeCurve;
import jp.sagalab.jftk.curve.primitive.Ellipse;
import jp.sagalab.jftk.curve.primitive.EllipticArc;
import jp.sagalab.jftk.curve.primitive.Line;
import jp.sagalab.jftk.curve.primitive.OpenFreeCurve;
import jp.sagalab.jftk.curve.primitive.PrimitiveCurve;
import jp.sagalab.jftk.recognition.BoundingBox;
import jp.sagalab.jftk.snapper.GridSpace;
import jp.sagalab.jftk.snapper.SnappingTargetData;
import jp.sagalab.jftk.transform.TransformMatrix;

/**
 * SKITPanelを補助するユーティリティクラス。
 */
class SKITPanelHelper {
	/** ボックス */
	private static class Box {
		/** 基点のx座標 */
		private final int x;
		/** 基点のy座標 */
		private final int y;
		/**	基点のz座標*/
		private final int z;
		/** ボックスの幅 */
		private final int width;
		/** ボックスの高さ */
		private final int height;
		/** ボックスの奥行き */
		private final int depth;

		private Box( int _x, int _y, int _z, int _width, int _height, int _depth ) {
			x = _x;
			y = _y;
			z = _z;
			width = _width;
			height = _height;
			depth = _depth;
		}

		public int x() {
			return x;
		}

		public int y() {
			return y;
		}

		public int z() {
			return z;
		}

		public int width() {
			return width;
		}

		public int height() {
			return height;
		}

		public int depth() {
			return depth;
		}
	}

	private SKITPanelHelper() {
		throw new UnsupportedOperationException( "Not create instance." );
	}

	/**
	 * 指定された点列を包含する単位グリッド上のボックス頂点を生成する。
	 *
	 * @param _vertices 点列
	 * @return 指定された点列を包含する単位グリッド上のボックス
	 */
	private static Box createUnitGridBox( Point[] _vertices ) {
		double minX = Double.POSITIVE_INFINITY;
		double maxX = Double.NEGATIVE_INFINITY;
		double minY = Double.POSITIVE_INFINITY;
		double maxY = Double.NEGATIVE_INFINITY;
		double minZ = Double.POSITIVE_INFINITY;
		double maxZ = Double.NEGATIVE_INFINITY;
		for ( Point vertex : _vertices ) {
			double x = vertex.x();
			double y = vertex.y();
			double z = vertex.z();
			if ( x < minX ) {
				minX = x;
			}
			if ( x > maxX ) {
				maxX = x;
			}
			if ( y < minY ) {
				minY = y;
			}
			if ( y > maxY ) {
				maxY = y;
			}
			if ( z < minZ ) {
				minZ = z;
			}
			if ( z > maxZ ) {
				maxZ = z;
			}
		}
		minX = Math.floor( minX );
		maxX = Math.ceil( maxX );
		minY = Math.floor( minY );
		maxY = Math.ceil( maxY );
		minZ = Math.floor( minZ );
		maxZ = Math.ceil( maxZ );

		return new Box( (int) minX, (int) minY, (int) minZ,
			(int) maxX - (int) minX,
			(int) maxY - (int) minY,
			(int) maxZ - (int) minZ );
	}

	/**
	 * グリッドを描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _gridSpace 基準グリッド空間
	 * @param _relativeGridResoltion 基準グリッドから何段階解像度をあげたグリッドを表示するか。
	 * @param _matViewToModel ビューポート空間からモデル空間への変換行列
	 * @param _width ビューポート空間の描画領域の幅
	 * @param _height ビューポート空間の描画領域の高さ
	 */
	static void drawBasedGrid( Graphics2D _g, GridSpace _basedGridSpace, int _relativeGridResoltion, TransformMatrix _matViewToModel, double _width, double _height ) {
		GridSpace gridSpace = _basedGridSpace;
		for ( int i = 0; i > _relativeGridResoltion; --i ) {
			gridSpace = gridSpace.downResolution();
		}
		for ( int i = 0; i < _relativeGridResoltion; ++i ) {
			gridSpace = gridSpace.upResolution();
		}
		TransformMatrix grid = gridSpace.grid();
		TransformMatrix matModelToView = _matViewToModel.inverse();
		// ビューポート空間の描画領域の角隅をモデル空間へ変換
		Point leftTop = Point.createXYZ( 0.0, 0.0, 0.0 ).transform( _matViewToModel );
		Point rightTop = Point.createXYZ( _width, 0.0, 0.0 ).transform( _matViewToModel );
		Point leftBottom = Point.createXYZ( 0.0, _height, 0.0 ).transform( _matViewToModel );
		Point rightBottom = Point.createXYZ( _width, _height, 0.0 ).transform( _matViewToModel );

		// モデル空間から単位グリッド空間への変換行列
		TransformMatrix invGrid = grid.inverse();

		// モデル空間上の描画領域の角隅を単位グリッド空間上にマッピング
		Point unitLeftTop = leftTop.transform( invGrid );
		Point unitRightTop = rightTop.transform( invGrid );
		Point unitLeftBottom = leftBottom.transform( invGrid );
		Point unitRightBottom = rightBottom.transform( invGrid );

		// 単位グリッド上のボックスを生成
		Box box = createUnitGridBox( new Point[]{ unitLeftTop, unitRightTop, unitLeftBottom, unitRightBottom } );
		// グリッド線の太さを設定
		double strokeWidth = 1.0;
		BasicStroke wideStroke;
		if ( _relativeGridResoltion > 0 ) {
			// 高解像度のグリッドは破線で表示
			wideStroke = new BasicStroke( (float) strokeWidth, BasicStroke.CAP_BUTT,
				//ミシン目
				//				BasicStroke.JOIN_MITER, 10.0f, new float[]{ 2.0f, 3.0f }, 0.0f );
				//中央十字
				BasicStroke.JOIN_MITER, 10.0f, new float[]{ (float) ( grid.scalalize() / 2.0 ), (float) ( grid.scalalize() * 3.0 / 2.0 ) }, (float) ( grid.scalalize() * 5 / 4.0 ) );
		} else {
			// 解像度の段階によってグリッド線の太さを変える。
			strokeWidth = Math.pow( 2, -_relativeGridResoltion * 0.8 );
			wideStroke = new BasicStroke( (float) strokeWidth );
		}
		_g.setStroke( wideStroke );
		// グリッド描画
		drawGrid( _g, box, grid, matModelToView );
		// 基準グリッドなら、グリッドの原点を描画する
		if ( _relativeGridResoltion == 0 ) {
			double span = grid.scalalize() * matModelToView.scalalize();
			Point origin = Point.createXYZ( grid.get( 0, 3 ), grid.get( 1, 3 ), grid.get( 2, 3 ) );
			origin = origin.transform( matModelToView );
			drawCircle( _g, origin, span / 2 );
		}
	}

	/**
	 * スナッピング後の点を包含するグリッドを描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _p スナッピング後の点
	 * @param _gridSpace グリッド空間
	 * @param _matModelToView ビューポート空間からモデル空間への変換行列
	 */
	static void drawGridBox( Graphics2D _g, Point _p, GridSpace _gridSpace, TransformMatrix _matModelToView ) {
		TransformMatrix grid = _gridSpace.grid();
		// グリッドの幅
		double span = grid.scalalize() / 2;
		// モデル空間から単位グリッド空間への変換行列
		TransformMatrix invGrid = grid.inverse();
		Point[] points = new Point[]{
			Point.createXYZ( _p.x() + span, _p.y(), _p.z() ).transform( invGrid ),
			Point.createXYZ( _p.x(), _p.y() + span, _p.z() ).transform( invGrid ),
			Point.createXYZ( _p.x() - span, _p.y(), _p.z() ).transform( invGrid ),
			Point.createXYZ( _p.x(), _p.y() - span, _p.z() ).transform( invGrid ), };

		Box box = createUnitGridBox( points );
		// グリッド線の太さを設定
		BasicStroke wideStroke = new BasicStroke( 1 );
		_g.setStroke( wideStroke );
		drawGrid( _g, box, grid, _matModelToView );
	}

	private static void drawGrid( Graphics2D _g, Box _box, TransformMatrix _grid, TransformMatrix _matModelToView ) {
		int startX = _box.x();
		int startY = _box.y();
		int startZ = _box.z();
		int endX = _box.x() + _box.width();
		int endY = _box.y() + _box.height();
		int endZ = _box.z() + _box.depth();

		// 単位グリッド空間上でいうところのx方向のグリッド線を描画
		for ( int j = startY; j <= endY; ++j ) {
			for ( int k = startZ; k <= endZ; ++k ) {
				Point start = Point.createXYZ( startX, j, k ).transform( _grid ).transform( _matModelToView );
				Point end = Point.createXYZ( endX, j, k ).transform( _grid ).transform( _matModelToView );
				_g.draw( new Line2D.Double( start.x(), start.y(), end.x(), end.y() ) );
			}
		}
		// 単位グリッド空間上でいうところのy方向のグリッド線を描画
		for ( int j = startX; j <= endX; ++j ) {
			for ( int k = startZ; k <= endZ; ++k ) {
				Point start = Point.createXYZ( j, startY, k ).transform( _grid ).transform( _matModelToView );
				Point end = Point.createXYZ( j, endY, k ).transform( _grid ).transform( _matModelToView );
				_g.draw( new Line2D.Double( start.x(), start.y(), end.x(), end.y() ) );
			}
		}
		// 単位グリッド空間上でいうところのz方向のグリッド線を描画
		for ( int j = startX; j <= endX; ++j ) {
			for ( int k = startY; k <= endY; ++k ) {
				Point start = Point.createXYZ( j, k, startZ ).transform( _grid ).transform( _matModelToView );
				Point end = Point.createXYZ( j, k, endZ ).transform( _grid ).transform( _matModelToView );
				_g.draw( new Line2D.Double( start.x(), start.y(), end.x(), end.y() ) );
			}
		}
	}

	/**
	 * フラグメントを描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _curve フラグメントの曲線
	 */
	static void drawFragment( Graphics2D _g, ParametricCurve _curve ) {
		int num = (int) Math.max( 2, Math.ceil( _curve.range().length() / 0.01 ) );
		Point[] evaluated = _curve.evaluateAll( num, EvaluationType.TIME );
		for ( Point p : evaluated ) {
			drawCircle( _g, p, p.fuzziness() );
		}
	}

	/**
	 * プリミティブ曲線を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _primitive プリミティブ曲線
	 */
	static void drawPrimitive( Graphics2D _g, PrimitiveCurve _primitive ) {
		// プリミティブ曲線の評価点列描画
		for ( BezierCurve bezier : _primitive.convert() ) {
			drawBezierCurve( _g, bezier );
		}
	}

	/**
	 * Bezier曲線を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _bezier Bezier曲線
	 */
	static void drawBezierCurve( Graphics2D _g, BezierCurve _bezier ) {
		Path2D.Double path = new Path2D.Double();
		Point[] cp = _bezier.controlPoints();
		path.moveTo( cp[0].x(), cp[0].y() );
		path.curveTo( cp[1].x(), cp[1].y(), cp[2].x(), cp[2].y(), cp[3].x(), cp[3].y() );
		_g.draw( path );
	}

	/**
	 * ファジィ点列を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _points ファジィ点列
	 */
	static void drawFuzzyPoints( Graphics2D _g, Point[] _points ) {
		for ( Point p : _points ) {
			drawCircle( _g, p, p.fuzziness() );
		}
	}

	/**
	 * 指定された曲線の矢尻を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _curve 曲線
	 */
	static void drawArrowHeadOfCurve( Graphics2D _g, PrimitiveCurve _curve ) {
		BezierCurve[] beziers = _curve.convert();
		for ( int i = 0; i < beziers.length; ++i ) {
			Point[] cp = beziers[beziers.length - i - 1].controlPoints();
			int degree = cp.length - 1;
			for ( int j = 0; j < degree; ++j ) {
				Point start = cp[cp.length - j - 2];
				Point end = cp[cp.length - j - 1];
				Vector vector = Vector.createSE( start, end );
				boolean result = SKITPanelHelper.fillArrawHead( _g, end, vector, 15.0, Math.PI / 4.0 );
				if ( result ) {
					return;
				}
			}
		}
	}

	/**
	 * 指定された点の位置に指定された向きの矢尻(二等辺三角形)を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _point 矢尻の位置
	 * @param _vector 矢尻の向き
	 * @param _length 矢尻の長さ
	 * @param _acuteAngle 矢尻の角度(二等辺三角形の頂角の角度)
	 * @return 描画に成功したならばtrue
	 */
	static boolean fillArrawHead( Graphics2D _g, Point _point, Vector _vector, double _length, double _acuteAngle ) {
		// 矢尻の形を決める頂点列を定義
		double halfLength = _length / 2.0;
		double halfWidth = _length * Math.tan( _acuteAngle / 2.0 );
		Point[] vertices = new Point[]{
			Point.createXYZ( -halfLength, -halfWidth, 0.0 ),
			Point.createXYZ( halfLength, 0.0, 0.0 ),
			Point.createXYZ( -halfLength, halfWidth, 0.0 )
		};
		// 変換行列を生成
		Vector initVector = Vector.createXYZ( 1.0, 0.0, 0.0 );
		double angle = initVector.angle( _vector );
		if ( !Double.isNaN( angle ) ) {
			if ( _vector.y() < 0.0 ) {
				angle *= -1.0;
			}
			TransformMatrix mat = TransformMatrix.rotation( angle, Vector.createXYZ( 0.0, 0.0, 1.0 ) );
			mat = mat.product( TransformMatrix.translation( _point.x(), _point.y(), _point.z() ) );
			// 頂点列を変換
			for ( int i = 0; i < vertices.length; ++i ) {
				vertices[i] = vertices[i].transform( mat );
			}
			// 矢尻を描画
			fillPolygon( _g, vertices );
			return true;
		}
		return false;
	}

	/**
	 * 円を塗りつぶす。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _center 円の中心点
	 * @param _radius 半径
	 */
	static void fillCircle( Graphics2D _g, Point _center, double _radius ) {
		double size = _radius * 2;
		double x = _center.x() - _radius;
		double y = _center.y() - _radius;
		Ellipse2D circle = new Ellipse2D.Double( x, y, size, size );
		_g.fill( circle );
	}

	/**
	 * 円の輪郭を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _center 円の中心点
	 * @param _radius 半径
	 */
	static void drawCircle( Graphics2D _g, Point _center, double _radius ) {
		double size = _radius * 2;
		double x = _center.x() - _radius;
		double y = _center.y() - _radius;
		Ellipse2D circle = new Ellipse2D.Double( x, y, size, size );
		_g.draw( circle );
	}

	/**
	 * 線分を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _pre 端点
	 * @param _post 端点
	 */
	static void drawLine( Graphics2D _g, Point _pre, Point _post ) {
		Line2D line = new Line2D.Double( _pre.x(), _pre.y(), _post.x(), _post.y() );
		_g.draw( line );
	}

	/**
	 * 多角形を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _vertices 頂点列
	 */
	private static void drawPolygon( Graphics2D _g, Point[] _vertices ) {
		Path2D.Double path = new Path2D.Double();
		path.moveTo( _vertices[0].x(), _vertices[0].y() );
		for ( int i = 1; i < _vertices.length; ++i ) {
			path.lineTo( _vertices[i].x(), _vertices[i].y() );
		}
		path.closePath();
		_g.draw( path );
	}

	/**
	 * 塗りつぶし多角形を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _vertices 頂点列
	 */
	static void fillPolygon( Graphics2D _g, Point[] _vertices ) {
		Path2D.Double path = new Path2D.Double();
		path.moveTo( _vertices[0].x(), _vertices[0].y() );
		for ( int i = 1; i < _vertices.length; ++i ) {
			path.lineTo( _vertices[i].x(), _vertices[i].y() );
		}
		path.closePath();
		_g.fill( path );
	}

	/**
	 * プリミティブ曲線の曲線種を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _shpae プリミティブ曲線
	 */
	static void drawPrimitiveType( Graphics2D _g, PrimitiveCurve _primitive, Font _font, Color _fgColor, Color _bgColor ) {
		float shadowDistance = _font.getSize2D() * 0.08f;
		Range range = _primitive.range();
		Point c = _primitive.locus( ( range.end() - range.start() ) / 2 + range.start() );
		String primitiveName;
		if ( _primitive instanceof Line ) {
			primitiveName = _primitive.isClosed() ? "P" : "L";
		} else if ( _primitive instanceof Circle ) {
			primitiveName = "C";
		} else if ( _primitive instanceof CircularArc ){
			primitiveName = "CA";
		} else if ( _primitive instanceof Ellipse ) {
			primitiveName = "E";
		} else if ( _primitive instanceof EllipticArc) {
			primitiveName = "EA";
		} else if ( _primitive instanceof ClosedFreeCurve) {
			primitiveName = "FC";
		} else if ( _primitive instanceof OpenFreeCurve) {
			primitiveName = "FO";
		} else {
			primitiveName = null;
		}
		_g.setFont( _font );
		_g.setPaint( _bgColor );
		_g.drawString( primitiveName, (float) c.x() + shadowDistance, (float) c.y() + shadowDistance );
		_g.setPaint( _fgColor );
		_g.drawString( primitiveName, (float) c.x(), (float) c.y() );
	}

	/**
	 * 文字列を描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _text 文字列
	 * @param _position 描画位置
	 * @param _font フォント
	 * @param _color 描画色
	 */
	static void drawText( Graphics2D _g, String _text, Point _position, Font _font, Color _color ) {
		_g.setFont( _font );
		_g.setPaint( _color );
		_g.drawString( _text, (float) _position.x(), (float) _position.y() );
	}

	/**
	 * ガイドラインを描画する。
	 *
	 * @param _g Graphics2Dオブジェクト。
	 * @param _primitive プリミティブ曲線。
	 * @param _snappingData スナッピングデータ。
	 * @param _matModelToView モデル空間からビューポート空間への変換行列。
	 */
	static void drawGuideline( Graphics2D _g, PrimitiveCurve _primitive,
		SnappingTargetData _snappingData, TransformMatrix _matModelToView ) {
		if ( _primitive instanceof Circle || _primitive instanceof CircularArc ) {
			drawCircularGuideLine( _g, (CircularArc) _primitive, _snappingData, _matModelToView );
		} else if ( _primitive instanceof Ellipse || _primitive instanceof EllipticArc ) {
			drawEllipticGuideLine( _g, (EllipticArc) _primitive, _snappingData, _matModelToView );
		}
	}

	/**
	 * 円弧のガイドラインを描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _circular 円弧
	 * @param _snappingData スナッピングデータ。
	 * @param _matModelToView モデル空間からビューポート空間への変換行列。
	 */
	private static void drawCircularGuideLine( Graphics2D _g, CircularArc _circular,
		SnappingTargetData _snappingData, TransformMatrix _matModelToView ) {

		BoundingBox boundingBox = _snappingData.getBoundingBox();
		if ( boundingBox != null ) {
			for ( int i = 0; i < boundingBox.getSmallRectLength(); ++i ) {
				Point[] miniBox = boundingBox.getSmallRectangle( i );
				for ( int j = 0; j < miniBox.length; ++j ) {
					miniBox[j] = miniBox[j].transform( _matModelToView );
				}
				drawPolygon( _g, miniBox );
			}
		}
	}

	/**
	 * 楕円弧のガイドラインを描画する。
	 *
	 * @param _g Graphics2Dオブジェクト
	 * @param _elliptic 楕円弧
	 * @param _snappingData スナッピングデータ。
	 * @param _matModelToView モデル空間からビューポート空間への変換行列。
	 */
	private static void drawEllipticGuideLine( Graphics2D _g, EllipticArc _elliptic,
		SnappingTargetData _snappingData, TransformMatrix _matModelToView ) {

		BoundingBox boundingBox = _snappingData.getBoundingBox();
		if ( boundingBox != null ) {
			for ( int i = 0; i < boundingBox.getSmallRectLength(); ++i ) {
				Point[] miniBox = boundingBox.getSmallRectangle( i );
				for ( int j = 0; j < miniBox.length; ++j ) {
					miniBox[j] = miniBox[j].transform( _matModelToView );
				}
				drawPolygon( _g, miniBox );
			}
		}
	}
}
