package jp.sagalab.skit;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import jp.sagalab.jflib.Shape;

/**
 * 図形データベース。
 *
 * このクラスのメソッドはマルチスレッド環境で実行しても適切に同期が行われる。
 */
public class ShapeDatabase {
	/** ホット図形群 */
	private final List<Shape> m_hots;
	/** アクティブ図形群 */
	private final List<Shape> m_actives;
	/** 非アクティブ図形群 */
	private final List<Shape> m_inactives;

	/**
	 * デフォルトの設定で生成したShapeDatabaseインスタンスを返す
	 * @return 図形データベース
	 */
	public static ShapeDatabase create() {
		return new ShapeDatabase( new LinkedList<Shape>(), new LinkedList<Shape>(), new LinkedList<Shape>() );
	}

	/**
	 * 図形群を指定して生成。
	 * @param _hots ホット図形群
	 * @param _actives アクティブ図形群
	 * @param _inactives 非アクティブ図形群
	 */
	protected ShapeDatabase( List<Shape> _hots, List<Shape> _actives, List<Shape> _inactives ) {
		m_hots = _hots;
		m_actives = _actives;
		m_inactives = _inactives;
	}

	/**
	 * この図形データベースの図形群に対して追加と削除を実行する。
	 * @param _add 追加する図形群。
	 * @param _remove 削除する図形群。
	 */
	void addAndRemoveActiveShapes( List<Shape> _add, List<Shape> _remove ) {
		synchronized ( this ) {
			m_hots.removeAll( _remove );
			m_actives.removeAll( _remove );
			m_actives.addAll( m_hots );
			m_hots.clear();
			m_hots.addAll( _add );
		}
	}

	/**
	 * この図形データベースの図形群に対して追加と削除を実行する。
	 * @param _add 追加する図形群。
	 * @param _remove 削除する図形群。
	 */
	void addAndRemoveInactiveShapes( List<Shape> _add, List<Shape> _remove ) {
		synchronized ( this ) {
			m_inactives.addAll( _add );
			m_inactives.removeAll( _remove );
		}
	}

	/**
	 * この図形データベースに登録されている図形群を返す。
	 * @return この図形データベースに登録されている図形群。
	 */
	public List<Shape> allShapes() {
		final List<Shape> allShapes;
		synchronized ( this ) {
			allShapes = new ArrayList<Shape>( m_hots.size() + m_actives.size() + m_inactives.size() );
			allShapes.addAll( m_inactives );
			allShapes.addAll( m_actives );
			allShapes.addAll( m_hots );
		}
		return allShapes;
	}

	/**
	 * この図形データベースに登録されているホット図形群とアクティブ図形群のリストを返す。
	 * @return ホット図形群とアクティブ図形群のリスト。
	 */
	public List<Shape> hotAndActiveShapes() {
		final List<Shape> shapes;
		synchronized ( this ) {
			shapes = new ArrayList<Shape>( m_actives.size() + m_inactives.size() );
			shapes.addAll( m_actives );
			shapes.addAll( m_hots );
		}
		return shapes;
	}

	/**
	 * この図形データベースに存在するホット図形群を返す。
	 * @return ホット図形群。
	 */
	public List<Shape> hotShapes() {
		synchronized ( this ) {
			return new ArrayList<Shape>( m_hots );
		}
	}

	/**
	 * この図形データベースに存在するアクティブ図形群を返す。
	 * @return アクティブ図形群。
	 */
	public List<Shape> activeShapes() {
		synchronized ( this ) {
			return new ArrayList<Shape>( m_actives );
		}
	}

	/**
	 * この図形データベースに存在する非アクティブ図形群を返す。
	 * @return 非アクティブ図形群。
	 */
	public List<Shape> inactiveShapes() {
		synchronized ( this ) {
			return new ArrayList<Shape>( m_inactives );
		}
	}
}
