package jp.sagalab.skit;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringReader;
import java.net.Socket;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import jp.sagalab.jflib.Shape;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.curve.BezierCurve;
import jp.sagalab.jftk.curve.Range;
import jp.sagalab.jftk.curve.primitive.Circle;
import jp.sagalab.jftk.curve.primitive.CircularArc;
import jp.sagalab.jftk.curve.primitive.ClosedFreeCurve;
import jp.sagalab.jftk.curve.primitive.Ellipse;
import jp.sagalab.jftk.curve.primitive.EllipticArc;
import jp.sagalab.jftk.curve.primitive.Line;
import jp.sagalab.jftk.curve.primitive.OpenFreeCurve;
import jp.sagalab.jftk.curve.primitive.PrimitiveCurve;
import jp.sagalab.jftk.transform.TransformMatrix;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * 作図アプリケーションと通信を行うためのクラス
 * @author hasegawa
 */
public class XMLServer {
	/** クライアントソケット */
	private final Socket m_clientSocket;

	public XMLServer( Socket _clientSocket ) {
		m_clientSocket = _clientSocket;
	}

	/**
	 * クライアントの接続を断つ。
	 * @throws IOException 処理の途中で入出力エラーが発生した場合。
	 * @throws IllegalStateException isConnected()がfalseを返す場合。
	 */
	public void close() throws IOException, IllegalStateException {
		// 接続を切る
		m_clientSocket.close();
	}

	/**
	 * クライアントへ図形プリミティブ情報を送信する。
	 * @param _shapes 図形群
	 * @throws IOException
	 */
	public void send( List<Shape> _shapes ) throws IOException, IllegalStateException {
		try {
			// 図形プリミティブ情報をXML文書化
			Document doc = createXMLDocument( _shapes );
			// XML文書送信
			TransformerFactory.newInstance().newTransformer().transform( new DOMSource( doc ), new StreamResult( m_clientSocket.getOutputStream() ) );
			PrintWriter writer = new PrintWriter( m_clientSocket.getOutputStream() );
			// データの区切りとして改行コードを送信
			writer.print( "\n" );
			writer.flush();
		} catch ( TransformerException ex ) {
			Logger.getLogger( XMLServer.class.getName() ).log( Level.SEVERE, null, ex );
		} catch ( ParserConfigurationException ex ) {
			Logger.getLogger( XMLServer.class.getName() ).log( Level.SEVERE, null, ex );
		}
	}

	/**
	 * 空のXMLデータを送信
	 */
	public void send() {
		PrintWriter writer = null;
		try {
			writer = new PrintWriter( m_clientSocket.getOutputStream() );
			writer.print( "<skit></skit>\n" );
			writer.flush();
		} catch ( IOException ex ) {
			Logger.getLogger( XMLServer.class.getName() ).log( Level.SEVERE, null, ex );
		} finally {
			if( writer != null){
				writer.close();
			}
		}
	}

	/**
	 * クライアントからビューポート情報とモデル空間情報を受信する
	 * @param _vi
	 * @param _mi
	 * @throws IOException
	 */
	public void receive( ViewportInfo _vi, ModelSpaceInfo _mi ) throws IOException {
		try {
			// クライアントからのデータを受信
			String str = new BufferedReader( new InputStreamReader( m_clientSocket.getInputStream() ) ).readLine();
			System.out.println( "ok!" );
			// 受信したXML文書を解析
			Document doc;
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			doc = builder.parse( new InputSource( new StringReader( str ) ) );
			// ビューポート情報とモデル空間情報を取得
			getViewportInfo( doc, _vi );
			getModelSpaceInfo( doc, _mi );
		} catch ( SAXException ex ) {
			Logger.getLogger( XMLServer.class.getName() ).log( Level.SEVERE, null, ex );
		} catch ( ParserConfigurationException ex ) {
			Logger.getLogger( XMLServer.class.getName() ).log( Level.SEVERE, null, ex );
		}
	}

	/**
	 * XML文書作成
	 * @return
	 */
	private Document createXMLDocument( List<Shape> _shapes ) throws ParserConfigurationException {
		// Documentオブジェクト生成
		Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		// ルート要素を生成
		Element e_skit = doc.createElement( "skit" );
		int shapeListSize = _shapes.size();
		// shapes要素を生成
		Element e_shapes = doc.createElement( "shapes" );
		// shapes要素の属性に図形数を付加
		e_shapes.setAttribute( "length", String.valueOf( shapeListSize ) );
		// プリミティブ列の情報をXML文書化
		for ( int i = 0; i < shapeListSize; ++i ) {
			// プリミティブ列
			List<PrimitiveCurve> primList = _shapes.get( i ).primitives();
			int primListSize = primList.size();
			// shape要素を生成
			Element e_shape = doc.createElement( "shape" );
			// TODO 閉じ判定をちゃんとする
			// shape要素の属性に図形の閉じ判定を付加
			if ( primListSize == 1 ) {
				e_shape.setAttribute( "closed_state", String.valueOf( primList.get( 0 ).isClosed() ) );
			} else {
				e_shape.setAttribute( "closed_state", String.valueOf( false ) );
			}
			// shape要素の属性に図形を形成するプリミティブ数を付加
			e_shape.setAttribute( "length", String.valueOf( primListSize ) );
			// 各プリミティブをXML文書に変換
			for ( int j = 0; j < primListSize; ++j ) {
				// プリミティブ要素
				Element e_prim = null;
				// プリミティブ
				PrimitiveCurve prim = primList.get( j );
				if ( prim instanceof Line ) {
					// 線分要素生成
					e_prim = createLineElement( doc, (Line) prim );
				} else if ( prim instanceof Circle || prim instanceof CircularArc ) {
					// 円要素生成
					e_prim = createCircularElement( doc, (CircularArc) prim );
				} else if ( prim instanceof Ellipse || prim instanceof EllipticArc ) {
					// 楕円要素生成
					e_prim = createEllipticElement( doc, (EllipticArc) prim );
				} else if ( prim instanceof OpenFreeCurve || prim instanceof ClosedFreeCurve ) {
					// 自由曲線要素生成
					e_prim = createFreeCurveElement( doc, (OpenFreeCurve) prim );
				}
				// shape要素の子要素に設定
				e_shape.appendChild( e_prim );
			}
			// shapes要素の子要素に設定
			e_shapes.appendChild( e_shape );
		}
		// ルート要素の子要素に設定
		e_skit.appendChild( e_shapes );
		doc.appendChild( e_skit );
		return doc;
	}

	/**
	 * XML文書からビューポート情報を取得するメソッド
	 * @param _doc
	 * @param _vi
	 * @return
	 */
	private void getViewportInfo( Document _doc, ViewportInfo _vi ) {
		// ノードの取得
		NodeList viList = _doc.getElementsByTagName( "viewport_information" );
		// 要素の取得
		Element e_vi = (Element) viList.item( 0 );
		// 属性値の取得
		_vi.setX( Integer.parseInt( e_vi.getAttribute( "x" ) ) );
		_vi.setY( Integer.parseInt( e_vi.getAttribute( "y" ) ) );
		_vi.setWidth( Integer.parseInt( e_vi.getAttribute( "width" ) ) );
		_vi.setHeight( Integer.parseInt( e_vi.getAttribute( "height" ) ) );
	}

	/**
	 * XML文書からモデル空間情報を取得するメソッド
	 * @param _doc
	 * @param _mi
	 * @return
	 */
	private void getModelSpaceInfo( Document _doc, ModelSpaceInfo _mi ) {
		// 変換行列ノードの取得
		NodeList mappingList = _doc.getElementsByTagName( "viewport_space_to_model_space_mapping_matrix" );
		// 要素の取得
		Element e_mapping = (Element) mappingList.item( 0 );
		// 属性値の取得
		double[][] mappingElements = new double[ 4 ][ 4 ];
		for ( int i = 0; i < 4; ++i ) {
			for ( int j = 0; j < 4; ++j ) {
				mappingElements[i][j] = Double.parseDouble( e_mapping.getAttribute( "m" + i + j ) );
			}
		}
		// 変換行列の生成
		TransformMatrix mappingMatrix = TransformMatrix.create(mappingElements );
		_mi.setMappingMatrix( mappingMatrix );

		// グリッド空間行列ノードの取得
		NodeList gridList = _doc.getElementsByTagName( "unit_grid_space_to_model_space_mapping_matrix" );
		// 要素の取得
		Element e_grid = (Element) gridList.item( 0 );
		// 属性値の取得
		double[][] gridElements = new double[ 4 ][ 4 ];
		for ( int j = 0; j < 4; ++j ) {
			for ( int k = 0; k < 4; ++k ) {
				gridElements[j][k] = Double.parseDouble( e_grid.getAttribute( "m" + j + k ) );
			}
		}
		TransformMatrix gridMatrix = TransformMatrix.create(gridElements );
		_mi.setGridSpaceMatrix( gridMatrix );

		// 派生倍率の取得
		NodeList dsfList = _doc.getElementsByTagName( "derivational_scale_factor" );
		Element e_dsf = (Element) dsfList.item( 0 );
		int derivatinalScaleFactor = Integer.parseInt( e_dsf.getTextContent() );
		_mi.setDerivatinalScaleFactor( derivatinalScaleFactor );
	}

	/**
	 * 線分のDOM要素構築
	 * @param _doc 構築先ドキュメント
	 * @param _line 線分プリミティブ
	 * @return 入力された線分に対するDOM要素
	 */
	private Element createLineElement( Document _doc, Line _line ) {
		// 線分要素生成
		Element e_line = _doc.createElement( "line" );
		// 始点座標の記述
		e_line.appendChild( createPointElement( _doc, _line.start(), "start" ) );
		// 終点座標の記述
		e_line.appendChild( createPointElement( _doc, _line.end(), "end" ) );
		return e_line;
	}

	/**
	 * 点のDOM要素構築
	 * @param _doc 構築先ドキュメント
	 * @param _point 点
	 * @param _tagName タグ名
	 * @return 入力された点に対するDOM要素
	 */
	private Node createPointElement( Document _doc, Point _point, String _tagName ) {
		// 点要素生成
		Element e_point = _doc.createElement( _tagName );
		// 属性値を設定
		e_point.setAttribute( "x", Double.toString( _point.x() ) );
		e_point.setAttribute( "y", Double.toString( _point.y() ) );
		e_point.setAttribute( "z", Double.toString( _point.z() ) );
		return e_point;
	}

	/**
	 * 円のDOM要素構築
	 * @param _doc 構築先ドキュメント
	 * @param _circular 円プリミティブ
	 * @return 入力された点に対するDOM要素
	 */
	private Element createCircularElement( Document _doc, CircularArc _circular ) {
		// 円要素生成
		Element e_circle = _doc.createElement( _circular.isClosed() ? "circle" : "circular_arc" );
		// 中心座標の記述
		e_circle.appendChild( createPointElement( _doc, _circular.center(), "center" ) );
		// 半径の記述
		Element e_radius = _doc.createElement( "radius" );
		e_radius.appendChild( _doc.createTextNode( Double.toString( _circular.radius() ) ) );
		e_circle.appendChild( e_radius );
		// 存在範囲の記述
		Range angleRange = _circular.range();
		Element e_angle = _doc.createElement( "angle" );
		e_angle.setAttribute( "start", Double.toString( angleRange.start() ) );
		e_angle.setAttribute( "end", Double.toString( angleRange.end() ) );
		e_circle.appendChild( e_angle );
		// 姿勢行列の記述
		TransformMatrix postureMat = _circular.posture();
		Element e_posture = _doc.createElement( "posture" );
		double[][] elements = postureMat.elements();
		for ( int i = 0; i < elements.length; ++i ) {
			for ( int j = 0; j < elements[i].length; j++ ) {
				e_posture.setAttribute( "m" + i + j, Double.toString( postureMat.get( i, j ) ) );
			}
		}
		e_circle.appendChild( e_posture );
		return e_circle;
	}

	/**
	 * 楕円のDOM要素の構築
	 * @param _doc 構築先ドキュメント
	 * @param _elliptic 楕円プリミティブ
	 * @return 入力された楕円に対応するDOM要素
	 */
	private Element createEllipticElement( Document _doc, EllipticArc _elliptic ) {
		// 楕円要素生成
		Element e_elliptic = _doc.createElement( _elliptic.isClosed() ? "ellipse" : "elliptic_arc" );
		// 中心座標の記述
		e_elliptic.appendChild( createPointElement( _doc, _elliptic.center(), "center" ) );
		// 長径の記述
		Element e_majorRadius = _doc.createElement( "major_radius" );
		e_majorRadius.appendChild( _doc.createTextNode( Double.toString( _elliptic.majorRadius() ) ) );
		e_elliptic.appendChild( e_majorRadius );
		// 短径の記述
		Element e_minorRadius = _doc.createElement( "minor_radius" );
		e_minorRadius.appendChild( _doc.createTextNode( Double.toString( _elliptic.minorRadius() ) ) );
		e_elliptic.appendChild( e_minorRadius );
		// 存在範囲の記述
		Range angleRange = _elliptic.range();
		Element e_angle = _doc.createElement( "angle" );
		e_angle.setAttribute( "start", Double.toString( angleRange.start() ) );
		e_angle.setAttribute( "end", Double.toString( angleRange.end() ) );
		e_elliptic.appendChild( e_angle );
		// 姿勢行列の記述
		TransformMatrix postureMat = _elliptic.posture();
		Element e_posture = _doc.createElement( "posture" );
		double[][] elements = postureMat.elements();
		for ( int i = 0; i < elements.length; ++i ) {
			for ( int j = 0; j < elements[i].length; ++j ) {
				e_posture.setAttribute( "m" + i + j, Double.toString( postureMat.get( i, j ) ) );
			}
		}
		e_elliptic.appendChild( e_posture );
		return e_elliptic;
	}

	/**
	 * 自由曲線のDOM要素の構築
	 * @param _doc 構築先ドキュメント
	 * @param _freeCurve 自由曲線プリミティブ
	 * @return 入力された自由曲線に対するDOM要素
	 */
	private Element createFreeCurveElement( Document _doc, OpenFreeCurve _freeCurve ) {
		BezierCurve[] beziers = _freeCurve.beziers();
		// 自由曲線要素の生成
		Element e_freeCurve = _doc.createElement( _freeCurve.isClosed() ? "open_free_curve" : "closed_free_curve" );
		// 自由曲線要素の属性に次数を付加
		e_freeCurve.setAttribute( "degree", String.valueOf( _freeCurve.degree() ) );
		// 自由曲線要素の属性に自由曲線を形成するBezier曲線の数を付加
		e_freeCurve.setAttribute( "length", String.valueOf( beziers.length ) );
		// Bezier曲線列の記述
		for ( BezierCurve b : beziers ) {
			Point[] controlPoints = b.controlPoints();
			// Bezier曲線要素の生成
			Element e_bezierCurve = _doc.createElement( "bezier_curve" );
			// Bezier曲線要素の属性に制御点数を付加
			e_bezierCurve.setAttribute( "length", String.valueOf( controlPoints.length ) );
			// 制御点列の記述
			for ( Point p : controlPoints ) {
				e_bezierCurve.appendChild( createPointElement( _doc, p, "control_point" ) );
			}
			e_freeCurve.appendChild( e_bezierCurve );
		}
		return e_freeCurve;
	}
}
