package jp.sagalab.skit.mulchitouchdevice;

/**
 * デバイスの状態に変化があったことを示すイベント。
 * @author nagase
 */
public class MTDeviceEvent {
	/** デバイスイベントタイプ */
	public static enum DeviceEventType {
		/** ボタンプレス */
		FINGER_PRESSED,
		/** ボタンリリース */
		FINGER_RELEASED,
		/** デバイスムーブ */
		FINGER_MOVED,
	}

	/** 指ID */
	private final int m_fingerID;

	/** デバイス座標値 */
	private final double[] m_position;

	/** イベント発生時刻 */
	private final long m_time;

	/** イベントタイプ */
	private final DeviceEventType m_type;

	/**
	 * ボタンID、デバイス座標値、デバイス姿勢、イベント発生時刻、イベントタイプを指定するコンストラクタ。
	 * @param buttonID ボタンID。
	 * @param position デバイス座標値。
	 * @param posture  デバイス姿勢。
	 * @param time     イベント発生時刻。
	 * @param type     イベントタイプ。
	 */
	private MTDeviceEvent( int buttonID, double[] position, long time, DeviceEventType type ) {
		this.m_fingerID = buttonID;
		this.m_position = position;
		this.m_time = time;
		this.m_type = type;
	}

	/**
	 * ボタンプレスイベントの生成。
	 * @param fingerID ボタンID。
	 * @param position 
	 * @param time     イベント発生時刻。
	 * @return ボタンプレスイベント。
	 */
	public static MTDeviceEvent createPressedEvent( int fingerID, double[] position, long time ) {
		return new MTDeviceEvent( fingerID, position, time, DeviceEventType.FINGER_PRESSED );
	}

	/**
	 * ボタンリリースイベントの生成。
	 * @param fingerID ボタンID。
	 * @param position
	 * @param time     イベント発生時刻。
	 * @return ボタンリリースイベント。
	 */
	public static MTDeviceEvent createReleasedEvent( int fingerID, double[] position, long time ) {
		return new MTDeviceEvent( fingerID, position, time, DeviceEventType.FINGER_RELEASED );
	}

	/**
	 * デバイスムーブイベントの生成。
	 * @param fingerID
	 * @param position デバイス座標値。
	 * @param time     イベント発生時刻。
	 * @return デバイスムーブイベント。
	 */
	public static MTDeviceEvent createMovedEvent( int fingerID, double[] position, long time ) {
		return new MTDeviceEvent( fingerID, position, time, DeviceEventType.FINGER_MOVED );
	}

	/**
	 * 指IDの取得。
	 * @return ボタンID。
	 */
	public int getFingerID() {
		return m_fingerID;
	}

	/**
	 * デバイスのx座標値の取得。
	 * @return デバイスのx座標値。
	 */
	public double getX() {
		if ( m_position != null ) {
			return m_position[0];
		}
		return 0;
	}

	/**
	 * デバイスのy座標値の取得。
	 * @return デバイスのy座標値。
	 */
	public double getY() {
		if ( m_position != null ) {
			return m_position[1];
		}
		return 0;
	}

	/**
	 * デバイスのz座標値の取得。
	 * @return デバイスのz座標値。
	 */
	public double getZ() {
		if ( m_position != null ) {
			return m_position[2];
		}
		return 0;
	}

	/**
	 * イベント発生時刻の取得。
	 * @return イベント発生時刻。
	 */
	public long getTime() {
		return m_time;
	}

	/**
	 * イベントタイプの取得。
	 * @return イベントタイプ。
	 */
	public DeviceEventType getType() {
		return m_type;
	}
}
