package jp.sagalab.skit.mulchitouchdevice;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.ConnectException;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;
import jp.sagalab.skit.ViewportInfo;

/**
 * デバイスサーバに接続し、サーバから受信したデバイスイベントをリスナー群に配信するリモートデバイス。
 * @author nagase
 */
public class MultiTouchDeviceClient {
	/** ソケット */
	private Socket socket;
	/** ビューポート情報 */
	private final ViewportInfo vi;
	/** リスナー群 */
	private final CopyOnWriteArrayList<MTDeviceListener> listeners = new CopyOnWriteArrayList<MTDeviceListener>();

	public MultiTouchDeviceClient( ViewportInfo _vi ) {
		vi = _vi;
	}

	/**
	 * デバイスサーバへ接続。
	 *
	 * @param serverName デバイスサーバのホスト名。
	 * @param portNumber デバイスサーバのポート番号。
	 * @param id デバイス番号。
	 * @return デバイスサーバへ接続できた場合はtrue。
	 */
	public boolean connect( String serverName, int portNumber, final int id ) {
		// 既に接続済みであれば、先に切断しておく
		disconnect();

		try {
			// デバイスサーバに接続
			socket = new Socket( serverName, portNumber );

			// 受信用スレッドの開始
			Thread receiveTherad = new Thread() {
				@Override
				public void run() {
					Socket socket = MultiTouchDeviceClient.this.socket;
					try {
						// デバイスIDの送信
						DataOutputStream output = new DataOutputStream( socket.getOutputStream() );
						output.writeByte( id );

						// デバイス情報の受信
						DataInputStream input = new DataInputStream( socket.getInputStream() );
						// ソケットが閉じられるまで受信し続ける
						while ( true ) {
							// イベントタイプの受信
							int eventType = input.readUnsignedByte();
							// タッチされた指ごとに固有のIDを受信
							int fingerID = input.readUnsignedByte();
							// 指の座標値を受信。画面左上を起点としているのでビューポート空間(ウィンドウ)の左上起点になおす。
							double[] position = new double[]{ input.readDouble() - vi.X(), input.readDouble() - vi.Y(), input.readDouble() };
							// イベント発生時刻の受信
							long time = input.readLong();
							// イベントタイプでEventの生成を振り分け
							MTDeviceEvent event = null;
							switch ( eventType ) {
								case 0:
									event = MTDeviceEvent.createPressedEvent( fingerID, position, time );
									break;
								case 1:
									event = MTDeviceEvent.createReleasedEvent( fingerID, position, time );
									break;
								case 2:
									event = MTDeviceEvent.createMovedEvent( fingerID, position, time );
									break;
								default:
									break;
							}
							// イベントをリスナー群に配信
							processDeviceEvent( event );
						}
					}catch ( SocketException e ) {
						System.out.println( "MultiTouch Device Client was disconnected." );
					}catch ( IOException e ) {
						Logger.getLogger( MultiTouchDeviceClient.class.getName() ).log( Level.SEVERE, null, e );
						// 入出力エラー
						if ( socket != null ) {
							try {
								socket.close();
							} catch ( IOException ex ) {
								Logger.getLogger( MultiTouchDeviceClient.class.getName() ).log( Level.SEVERE, null, ex );
							}
						}
					}
				}
			};
			receiveTherad.start();
		} catch ( ConnectException e ) {
			// 接続失敗
			//Logger.getLogger(MulchiTouchDeviceClient.class.getName()).log(Level.SEVERE, null, e);
			return false;
		} catch ( UnknownHostException e ) {
			// ホストが見付からない
			Logger.getLogger( MultiTouchDeviceClient.class.getName() ).log( Level.SEVERE, null, e );
			return false;
		} catch ( IOException e ) {
			// 入出力エラー
			Logger.getLogger( MultiTouchDeviceClient.class.getName() ).log( Level.SEVERE, null, e );
			return false;
		}

		return true;
	}

	/**
	 * デバイスサーバとの接続の切断。
	 */
	public void disconnect() {
		if ( socket != null ) {
			try {
				socket.close();
			} catch ( IOException ex ) {
				Logger.getLogger( MultiTouchDeviceClient.class.getName() ).log( Level.SEVERE, null, ex );
			}
			socket = null;
		}
	}

	/**
	 * デバイスイベントリスナーの登録。
	 *
	 * @param listener デバイスイベントリスナー。
	 */
	public void addDeviceListener( MTDeviceListener listener ) {
		// 登録済みかをチェック
		int index = listeners.indexOf( listener );
		if ( index < 0 ) {
			// 既存のリスナーでなければ登録
			listeners.add( listener );
		}
	}

	/**
	 * デバイスイベントリスナーの削除。
	 *
	 * @param listener デバイスイベントリスナー。
	 */
	public void removeDeviceListener( MTDeviceListener listener ) {
		// リスナー群から削除
		listeners.remove( listener );
	}

	/**
	 * デバイスサーバとの接続状態の取得。
	 *
	 * @return 接続済みの場合はtrue。
	 */
	public boolean isConnected() {
		return socket != null;
	}

	/**
	 * デバイスイベントリスナー群へのデバイスイベントの配信。
	 *
	 * @param event デバイスイベント。
	 */
	protected void processDeviceEvent( MTDeviceEvent event ) {
		if ( event != null ) {
			List<MTDeviceListener> listeners = new ArrayList<MTDeviceListener>( this.listeners );
			for ( MTDeviceListener listener : listeners ) {
				switch ( event.getType() ) {
					case FINGER_PRESSED:
						listener.devicePressed( event );
						break;
					case FINGER_RELEASED:
						listener.deviceReleased( event );
						break;
					case FINGER_MOVED:
						listener.deviceMoved( event );
						break;
					default:
						System.out.println( "Unknown event" );
						break;
				}
			}
		}
	}
}
