package jp.sagalab.skit.mulchitouchdevice;

import java.awt.Container;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.SortedMap;
import java.util.TreeMap;
import jp.sagalab.jftk.Point;
import jp.sagalab.jftk.transform.TransformMatrix;
import jp.sagalab.skit.AbsoluteModelSpaceController;
import jp.sagalab.skit.RelativeModelSpaceController;
import jp.sagalab.skit.SKITModel;
import jp.sagalab.skit.SKITPanel;

/**
 * マルチタッチデバイスによる図形空間操作を行うクラス。
 * @author nagase
 */
public class MultiTouchModelSpaceController implements MTDeviceListener {
	
	@Override
	public synchronized void devicePressed( MTDeviceEvent e ) {
		// 指が触れている間、マウスの動きを封じる
		retractInputListener();
		// DeviceEventの情報からPointを生成
		Point current = Point.createXYZT( e.getX(), e.getY(), e.getZ(), e.getTime() );
		// 触れた指を管理対象に追加。
		m_fingers.put( e.getFingerID(), current );

		// 触れている指の本数によって場合わけ
		// 今は特になにもしてない。
		switch ( m_fingers.size() ) {
			case 1:
				break;
			case 2:
				break;
			case 3:
				break;
			default:
				break;
		}

	}

	@Override
	public synchronized void deviceMoved( MTDeviceEvent e ) {
		// PRESSイベントが来ずにいきなりMOVEDがくることがある。その場合の回避。
		if ( !m_fingers.containsKey( e.getFingerID() ) ) {
			devicePressed( e );
		}
		Point current = Point.createXYZT( e.getX(), e.getY(), e.getZ(), e.getTime() );
		Point last = m_fingers.get( e.getFingerID() );
		// 触れている指の本数で場合分け
		switch ( m_fingers.size() ) {
			case 1:
				// 相対的図形空間操作
				m_RMC.move( last.x() - current.x(), last.y() - current.y(), 0 );
				break;
			case 2:
				// 2本のうち、他方の指の取得
				Point pibot=null;
				for ( int i : m_fingers.keySet() ) {
					// 2本しかないはずなのでe.getFingerIDと違えばそれが他方の指ID。
					if ( i != e.getFingerID() ) {
						pibot=m_fingers.get( i );
					}
				}
				TransformMatrix mat = m_model.getViewToModel();
				// 絶対的図形空間操作
				m_AMC.move( pibot.transform( mat ), last.transform( mat ), current.transform( mat ) );
				break;
			default:
				break;
		}
		m_fingers.put( e.getFingerID(), current );

	}

	@Override
	public synchronized void deviceReleased( MTDeviceEvent e ) {
		// PRESSイベントが来ずにいきなりRELEASEDがくることがある。その場合の回避。
		if ( !m_fingers.containsKey( e.getFingerID() ) ) {
			devicePressed( e );
		}
		Point current = Point.createXYZT( e.getX(), e.getY(), e.getZ(), e.getTime() );
		Point last = m_fingers.get( e.getFingerID() );

		// 触れている指の本数で場合分け
		switch ( m_fingers.size() ) {
			case 1:
				// 相対的図形空間操作
				m_RMC.move( last.x() - current.x(), last.y() - current.y(), 0 );
				break;
			case 2:
				// 2本のうち、他方の指の取得
				Point pibot=null;
				for ( int i : m_fingers.keySet() ) {
					// 2本しかないはずなのでe.getFingerIDと違えばそれが他方の指ID。
					if ( i != e.getFingerID() ) {
						pibot=m_fingers.get( i );
					}
				}
				TransformMatrix mat = m_model.getViewToModel();
				// 絶対的図形空間操作
				m_AMC.move( pibot.transform( mat ), last.transform( mat ), current.transform( mat ) );
				break;
			default:
				break;
		}

		// 離した指を管理対象から削除
		m_fingers.remove( e.getFingerID() );
		
		// 触れている指が0本なら通常のリスナに戻す。
		if ( m_fingers.isEmpty() ) {
			// 離した瞬間OS経由でctrlのキーイベントが飛んでくることがある。
			// ピンチアウトなど特定のジェスチャにOSが反応しているようだ。
			// とりあえず数瞬止めておけばやり過ごせるという対処。ナガセ
			try {
				Thread.sleep( 100);
			} catch ( InterruptedException ex ) {
			}
			undoInputListener();
		} else {
			retractInputListener();
		}
	}

	/**
	 * {@link #retractInputListener() }で退避させたリスナをもとに戻す。
	 */
	private void undoInputListener() {
		if ( isNoInput ) {
			MouseListener[] mouseListeners = m_container.getMouseListeners();
			MouseMotionListener[] mouseMotionListeners = m_container.getMouseMotionListeners();
			KeyListener[] keyListeners = m_container.getKeyListeners();
			for ( MouseListener mouseListener : mouseListeners ) {
				m_container.removeMouseListener( mouseListener );
			}
			for ( MouseMotionListener mouseMotionListener : mouseMotionListeners ) {
				m_container.removeMouseMotionListener( mouseMotionListener );
			}
			for ( KeyListener keyListener : keyListeners ) {
				m_container.removeKeyListener( keyListener );
			}

			for ( MouseListener mouseListener : tmpML ) {
				m_container.addMouseListener( mouseListener );
			}
			for ( MouseMotionListener mouseMotionListener : tmpMML ) {
				m_container.addMouseMotionListener( mouseMotionListener );
			}
			for ( KeyListener keyListener : tmpKL ) {
				m_container.addKeyListener( keyListener );
			}
		}
		isNoInput = false;
	}

	/**
	 * 標準のリスナを一時的に退避させ、入力を無効化する。
	 * 指が触れている間、曲線入力されないようにするために使われる。
	 */
	private void retractInputListener() {
		if ( !isNoInput ) {
			tmpML = m_container.getMouseListeners();
			tmpMML = m_container.getMouseMotionListeners();
			tmpKL = m_container.getKeyListeners();
			for ( MouseListener mouseListener : tmpML ) {
				m_container.removeMouseListener( mouseListener );
			}
			for ( MouseMotionListener mouseMotionListener : tmpMML ) {
				m_container.removeMouseMotionListener( mouseMotionListener );
			}
			for ( KeyListener keyListener : tmpKL ) {
				m_container.removeKeyListener( keyListener );
			}
		}
		isNoInput = true;
	}
	/** コンテナ */
	private final Container m_container;
	/** 通常のマウスリスナの退避用 */
	private MouseListener[] tmpML;
	/** 通常のマウスモーションリスナの退避用 */
	private MouseMotionListener[] tmpMML;
	/** 通常のキーリスナの退避用 */
	private KeyListener[] tmpKL;
	/** 通常のマウスリスナの退避さえているかどうかを表す */
	private boolean isNoInput = false;
	/** モデル */
	private final SKITModel m_model;
	/** 相対的図形空間操作を行うコントローラ */
	private final RelativeModelSpaceController m_RMC;
	/** 絶対的図形空間操作を行うコントローラ */
	private final AbsoluteModelSpaceController m_AMC;
	/** 触れている指を管理するマップ */
	private final SortedMap<Integer, Point> m_fingers;

	/**
	 * マルチタッチ図形空間操作コントローラを生成する。
	 * @param _model モデル
	 * @param _container パネル
	 * @param _relativeModelSpaceController 相対的図形空間操作コントローラ
	 */
	public MultiTouchModelSpaceController( SKITModel _model, SKITPanel _container, RelativeModelSpaceController _relativeModelSpaceController ) {
		m_model = _model;
		m_container = _container;
		m_RMC = _relativeModelSpaceController;
		m_AMC = AbsoluteModelSpaceController.create( _model );
		m_fingers = new TreeMap<Integer, Point>();
	}
}
